# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pytest

from snapcraft.internal import steps


def test_step_order():
    step = steps.PULL

    step = step.next_step()
    assert step == steps.BUILD

    step = step.next_step()
    assert step == steps.STAGE

    step = step.next_step()
    assert step == steps.PRIME

    assert step.next_step() is None

    step = step.previous_step()
    assert step == steps.STAGE

    step = step.previous_step()
    assert step == steps.BUILD

    step = step.previous_step()
    assert step == steps.PULL

    assert step.previous_step() is None


def test_next_step_handles_none():
    assert steps.next_step(None) == steps.PULL


@pytest.mark.parametrize(
    "step,next_steps",
    [
        (steps.PULL, [steps.BUILD, steps.STAGE, steps.PRIME]),
        (steps.BUILD, [steps.STAGE, steps.PRIME]),
        (steps.STAGE, [steps.PRIME]),
        (steps.PRIME, []),
    ],
)
def test_next_steps(step, next_steps):
    assert step.next_steps() == next_steps


@pytest.mark.parametrize(
    "step,previous_steps",
    [
        (steps.PULL, []),
        (steps.BUILD, [steps.PULL]),
        (steps.STAGE, [steps.PULL, steps.BUILD]),
        (steps.PRIME, [steps.PULL, steps.BUILD, steps.STAGE]),
    ],
)
def test_previous_steps(step, previous_steps):
    assert step.previous_steps() == previous_steps



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

import fixtures


class TestDir(fixtures.Fixture):
    def _setUp(self):
        self.path = self.useFixture(fixtures.TempDir()).path

    def create_file(self, *name: str) -> None:
        open(os.path.join(self.path, *name), "wb").close()

    def create_dir(self, *name: str) -> None:
        os.mkdir(os.path.join(self.path, *name))

    def unlink(self, *name) -> None:
        os.unlink(os.path.join(self.path, *name))

    def rmdir(self, *name: str) -> None:
        os.rmdir(os.path.join(self.path, *name))

    def exists(self, *parts) -> bool:
        return os.path.exists(os.path.join(self.path, *parts))


class Chdir:
    def __init__(self, path):
        self._path = os.path.expanduser(path)

    def __enter__(self):
        self._old_path = os.getcwd()
        os.chdir(self._path)

    def __exit__(self, etype, value, traceback):
        os.chdir(self._old_path)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from textwrap import dedent

from testscenarios import multiply_scenarios
from testtools.matchers import Equals

from snapcraft.extractors import ExtractedMetadata, _errors, setuppy
from tests import unit


class TestSetupPy:

    metadata = [
        (
            "description",
            dict(params=dict(version=None, description="test-description")),
        ),
        ("version", dict(params=dict(version="test-version", description=None))),
        (
            "key and version",
            dict(params=dict(description="test-description", version="test-version")),
        ),
    ]

    tools = [
        (
            "setuptools",
            dict(import_statement="import setuptools", method="setuptools.setup"),
        ),
        (
            "from setuptools",
            dict(import_statement="from setuptools import setup", method="setup"),
        ),
        (
            "distutils",
            dict(
                import_statement="import distutils.core", method="distutils.core.setup"
            ),
        ),
        (
            "from distutils.core",
            dict(import_statement="from distutils.core import setup", method="setup"),
        ),
    ]

    scenarios = multiply_scenarios(metadata, tools)

    def create_setuppy(self, import_statement, method, parameters):
        params = ['    {}="{}",'.format(k, v) for k, v in parameters.items() if v]

        fmt = dict(
            params="\n".join(params), import_statement=import_statement, method=method
        )

        with open("setup.py", "w") as setup_file:
            print(
                dedent(
                    """\
                {import_statement}

                {method}(
                    name='hello-world',
                {params}
                    author='Canonical LTD',
                    author_email='snapcraft@lists.snapcraft.io',
                )
            """
                ).format(**fmt),
                file=setup_file,
            )

    def test(self, tmp_work_path, import_statement, method, params):
        self.create_setuppy(import_statement, method, params)

        expected = ExtractedMetadata(**params)
        actual = setuppy.extract("setup.py", workdir=".")

        assert str(actual) == str(expected)
        assert actual == expected


class SetupPyErrorsTestCase(unit.TestCase):
    def test_unhandled_file_test_case(self):
        raised = self.assertRaises(
            _errors.UnhandledFileError, setuppy.extract, "unhandled-file", workdir="."
        )

        self.assertThat(raised.path, Equals("unhandled-file"))
        self.assertThat(raised.extractor_name, Equals("setup.py"))

    def test_bad_import(self):
        with open("setup.py", "w") as setup_file:
            print("import bad_module", file=setup_file)

        self.assertRaises(
            _errors.SetupPyImportError, setuppy.extract, "setup.py", workdir="."
        )

    def test_unsupported_setup(self):
        with open("setup.py", "w") as setup_file:
            print(
                dedent(
                    """\
                def setup(**kwargs):
                    # Raise what setuptools and distutils raise
                    raise SystemExit()

                setup(name='name', version='version')
                """
                ),
                file=setup_file,
            )

        self.assertRaises(
            _errors.SetupPyFileParseError, setuppy.extract, "setup.py", workdir="."
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from snapcraft.internal.errors import SnapcraftError


class DeltaGenerationError(SnapcraftError):
    """A delta failed to generate."""

    fmt = (
        "Could not generate {delta_format} delta.\n"
        "stdout log: {stdout_path}\n"
        "stdout: \n{stdout}\n"
        "---------"
        "stderr log: {stderr_path}\n"
        "stderr: \n{stderr}\n"
        "---------"
        "returncode: {returncode}"
    )


class DeltaGenerationTooBigError(SnapcraftError):
    """The generated delta was too large."""

    fmt = "delta saving is less than {delta_min_percentage}%."


class DeltaFormatOptionError(SnapcraftError):
    """A delta format option is not in the defined list."""

    fmt = (
        "delta_format must be a option in {format_options_list}.\n"
        "for now delta_format={delta_format!r}"
    )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os
import shutil
import subprocess
import sys

import requests

import snapcraft.internal.common
from snapcraft.internal.cache import FileCache
from snapcraft.internal.indicators import (
    download_requests_stream,
    download_urllib_source,
)

from . import errors
from ._checksum import split_checksum, verify_checksum


class Base:
    def __init__(
        self,
        source,
        source_dir,
        source_tag=None,
        source_commit=None,
        source_branch=None,
        source_depth=None,
        source_checksum=None,
        command=None,
    ):
        self.source = source
        self.source_dir = source_dir
        self.source_tag = source_tag
        self.source_commit = source_commit
        self.source_branch = source_branch
        self.source_depth = source_depth
        self.source_checksum = source_checksum
        self.source_details = None

        self.command = command
        self._checked = False

    def check(self, target: str):
        """Check if pulled sources have changed since target was created.

        :param str target: Path to target file.
        """
        self._checked = True
        return self._check(target)

    def update(self):
        """Update pulled source.

        :raises RuntimeError: If this function is called before `check()`.
        """
        if not self._checked:
            # This is programmer error
            raise RuntimeError("source must be checked before it's updated")
        self._update()

    def _check(self, target: str):
        """Check if pulled sources have changed since target was created.

        :param str target: Path to target file.
        """
        raise errors.SourceUpdateUnsupportedError(self)

    def _update(self):
        """Update pulled source."""
        raise errors.SourceUpdateUnsupportedError(self)

    def _run(self, command, **kwargs):
        try:
            subprocess.check_call(command, **kwargs)
        except subprocess.CalledProcessError as e:
            raise errors.SnapcraftPullError(command, e.returncode)

    def _run_output(self, command, **kwargs):
        try:
            return (
                subprocess.check_output(command, **kwargs)
                .decode(sys.getfilesystemencoding())
                .strip()
            )
        except subprocess.CalledProcessError as e:
            raise errors.SnapcraftPullError(command, e.returncode)


class FileBase(Base):
    def pull(self):
        source_file = None
        is_source_url = snapcraft.internal.common.isurl(self.source)

        # First check if it is a url and download and if not
        # it is probably locally referenced.
        if is_source_url:
            source_file = self.download()
        else:
            basename = os.path.basename(self.source)
            source_file = os.path.join(self.source_dir, basename)
            # We make this copy as the provisioning logic can delete
            # this file and we don't want that.
            try:
                shutil.copy2(self.source, source_file)
            except FileNotFoundError as exc:
                raise errors.SnapcraftSourceNotFoundError(self.source) from exc

        # Verify before provisioning
        if self.source_checksum:
            verify_checksum(self.source_checksum, source_file)

        # We finally provision, but we don't clean the target so override-pull
        # can actually have meaning when using these sources.
        self.provision(self.source_dir, src=source_file, clean_target=False)

    def download(self, filepath: str = None) -> str:
        if filepath is None:
            self.file = os.path.join(self.source_dir, os.path.basename(self.source))
        else:
            self.file = filepath

        # First check if we already have the source file cached.
        file_cache = FileCache()
        if self.source_checksum:
            algorithm, hash = split_checksum(self.source_checksum)
            cache_file = file_cache.get(algorithm=algorithm, hash=hash)
            if cache_file:
                # We make this copy as the provisioning logic can delete
                # this file and we don't want that.
                shutil.copy2(cache_file, self.file)
                return self.file

        # If not we download and store
        if snapcraft.internal.common.get_url_scheme(self.source) == "ftp":
            download_urllib_source(self.source, self.file)
        else:
            try:
                request = requests.get(self.source, stream=True, allow_redirects=True)
                request.raise_for_status()
            except requests.exceptions.RequestException as e:
                raise errors.SnapcraftRequestError(message=e)

            download_requests_stream(request, self.file)

        # We verify the file if source_checksum is defined
        # and we cache the file for future reuse.
        if self.source_checksum:
            algorithm, digest = verify_checksum(self.source_checksum, self.file)
            file_cache.cache(filename=self.file, algorithm=algorithm, hash=hash)
        return self.file



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from typing import Tuple

import lazr.restfulclient.errors
from launchpadlib.launchpad import Launchpad

from . import errors

logger = logging.getLogger(__name__)


def split_ppa_parts(*, ppa: str) -> Tuple[str, str]:
    ppa_split = ppa.split("/")
    if len(ppa_split) != 2:
        raise errors.AptPPAInstallError(ppa=ppa, reason="invalid PPA format")
    return ppa_split[0], ppa_split[1]


def get_launchpad_ppa_key_id(*, ppa: str) -> str:
    """Query Launchpad for PPA's key ID."""
    owner, name = split_ppa_parts(ppa=ppa)
    launchpad = Launchpad.login_anonymously("snapcraft", "production")
    launchpad_url = f"~{owner}/+archive/{name}"

    logger.debug(f"Loading launchpad url: {launchpad_url}")
    try:
        key_id = launchpad.load(launchpad_url).signing_key_fingerprint
    except lazr.restfulclient.errors.NotFound as error:
        raise errors.AptPPAInstallError(
            ppa=ppa, reason="not found on launchpad"
        ) from error

    logger.debug(f"Retrieved launchpad PPA key ID: {key_id}")
    return key_id



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import abc
import base64
import logging
import os
import pathlib
import platform
import shlex
import shutil
import sys
import tempfile
from textwrap import dedent
from typing import Any, Dict, Optional, Sequence

import pkg_resources
from xdg import BaseDirectory

import snapcraft
from snapcraft import yaml_utils
from snapcraft.internal import common, steps

from . import errors
from ._snap import SnapInjector

logger = logging.getLogger(__name__)


def _get_platform() -> str:
    return sys.platform


class Provider(abc.ABC):

    _INSTANCE_PROJECT_DIR = "~/project"

    def __init__(
        self,
        *,
        project,
        echoer,
        is_ephemeral: bool = False,
        build_provider_flags: Dict[str, str] = None,
    ) -> None:
        self.project = project
        self.echoer = echoer
        self._is_ephemeral = is_ephemeral

        self.instance_name = "snapcraft-{}".format(project._snap_meta.name)

        if project._snap_meta.version:
            self.snap_filename = "{}_{}_{}.snap".format(
                project._snap_meta.name, project._snap_meta.version, project.deb_arch
            )
        else:
            self.snap_filename = "{}_{}.snap".format(
                project._snap_meta.name, project.deb_arch
            )

        self.provider_project_dir = os.path.join(
            BaseDirectory.save_data_path("snapcraft"),
            "projects",
            project._snap_meta.name,
            self._get_provider_name(),
        )

        if build_provider_flags is None:
            build_provider_flags = dict()
        self.build_provider_flags = build_provider_flags.copy()

    def __enter__(self):
        try:
            self.create()
        except errors.ProviderBaseError:
            # Destroy is idempotent
            self.destroy()
            raise
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.destroy()

    @classmethod
    @abc.abstractclassmethod
    def ensure_provider(cls) -> None:
        """Necessary steps to ensure the provider is correctly setup."""

    @classmethod
    @abc.abstractclassmethod
    def setup_provider(cls, *, echoer) -> None:
        """Necessary steps to install the provider on the host."""

    @classmethod
    @abc.abstractclassmethod
    def _get_provider_name(cls) -> str:
        """Return the provider name."""

    @classmethod
    @abc.abstractclassmethod
    def _get_is_snap_injection_capable(cls) -> bool:
        """Return whether the provider can install snaps from the host."""

    @abc.abstractmethod
    def create(self) -> None:
        """Provider steps needed to create a fully functioning environment."""

    @abc.abstractmethod
    def destroy(self) -> None:
        """Provider steps needed to ensure the instance is destroyed.

        This method should be safe to call multiple times and do nothing
        if the instance to destroy is already destroyed.
        """

    @abc.abstractmethod
    def _run(
        self, command: Sequence[str], hide_output: bool = False
    ) -> Optional[bytes]:
        """Run a command on the instance."""

    @abc.abstractmethod
    def _launch(self):
        """Launch the instance."""

    @abc.abstractmethod
    def _start(self):
        """Start an existing the instance."""

    @abc.abstractmethod
    def _push_file(self, *, source: str, destination: str) -> None:
        """Push a file into the instance."""

    @abc.abstractmethod
    def _is_mounted(self, target: str) -> bool:
        """Query if there is a mount at target mount point."""

    @abc.abstractmethod
    def _mount(self, host_source: str, target: str) -> None:
        """Mount host source directory to target mount point."""

    def _mount_project(self) -> None:
        """Provider steps needed to make the project available to the instance.
        """
        target = (self._get_home_directory() / "project").as_posix()
        self._mount(self.project._project_dir, target)

        if self.build_provider_flags.get("SNAPCRAFT_BIND_SSH"):
            self._mount_ssh()

    def _mount_prime_directory(self) -> bool:
        """Mount the host prime directory into the provider.

        :returns: True if the prime directory was already mounted.
        """
        target = (self._get_home_directory() / "prime").as_posix()
        if self._is_mounted(target):
            return True

        self._mount(self.project.prime_dir, target)
        return False

    def _mount_ssh(self) -> None:
        """Mount ~/.ssh to target."""
        src = (pathlib.Path.home() / ".ssh").as_posix()
        target = (self._get_home_directory() / ".ssh").as_posix()
        self._mount(src, target)

    def expose_prime(self) -> None:
        """Provider steps needed to expose the prime directory to the host.
        """
        os.makedirs(self.project.prime_dir, exist_ok=True)
        if not self._mount_prime_directory():
            self._run(command=["snapcraft", "clean", "--unprime"])

    def execute_step(self, step: steps.Step) -> None:
        self._run(command=["snapcraft", step.name])

    def clean_parts(self, part_names: Sequence[str]) -> None:
        self._run(command=["snapcraft", "clean"] + list(part_names))

    def pack_project(self, *, output: Optional[str] = None) -> None:
        command = ["snapcraft", "snap"]
        if output:
            command.extend(["--output", output])
        self._run(command=command)

    def clean_project(self) -> bool:
        try:
            shutil.rmtree(self.provider_project_dir)
            return True
        except FileNotFoundError:
            return False

    @abc.abstractmethod
    def pull_file(self, name: str, destination: str, delete: bool = False) -> None:
        """
        Provider steps needed to retrieve a file from the instance, optionally
        deleting the source file after a successful retrieval.

        :param name: the remote filename.
        :type name: str
        :param destination: the local filename.
        :type destination: str
        :param delete: whether the file should be deleted.
        :type delete: bool
        """

    @abc.abstractmethod
    def shell(self) -> None:
        """Provider steps to provide a shell into the instance."""

    def launch_instance(self) -> None:
        # Clean project if we cannot trust existing environment, or it
        # no longer matches project target.
        if os.path.exists(self.provider_project_dir):
            self._ensure_compatible_build_environment()

        try:
            # An ProviderStartError exception here means we need to create
            self._start()
        except errors.ProviderInstanceNotFoundError:
            # If starting failed, we need to make sure our data directory is
            # clean
            if os.path.exists(self.provider_project_dir):
                shutil.rmtree(self.provider_project_dir)
            os.makedirs(self.provider_project_dir)
            # then launch
            self._launch()

            # Configure environment for snapcraft use.
            self._setup_environment()

            # Refresh repository caches.
            self._run(["apt-get", "update"])

            # And make sure we are using the latest from that cache.
            self._run(["apt-get", "dist-upgrade", "--yes"])

            # Install any packages that might be missing from the base
            # image, but may be required for snapcraft to function.
            self._run(["apt-get", "install", "--yes", "apt-transport-https"])

        # Always setup snapcraft after a start to bring it up to speed with
        # what is on the host.
        self._setup_snapcraft()

        # Always update snapd proxy settings to match current http(s) proxy
        # settings.
        self._setup_snapd_proxy()

        # Install any CA certificates requested by the user.
        certs_path = self.build_provider_flags.get("SNAPCRAFT_ADD_CA_CERTIFICATES")
        if certs_path:
            self._add_ca_certificates(pathlib.Path(certs_path))

    def _add_ca_certificates(self, certs_path: pathlib.Path) -> None:
        # Should have been validated by click already.
        if not certs_path.exists():
            raise RuntimeError(f"Unable to read CA certificates: {certs_path!r}")

        if certs_path.is_file():
            certificate_files = [certs_path]
        elif certs_path.is_dir():
            certificate_files = [x for x in certs_path.iterdir() if x.is_file()]
        else:
            raise RuntimeError(
                f"Unable to read CA certificates: {certs_path!r} (unhandled file type)"
            )

        for certificate_file in sorted(certificate_files):
            logger.info(f"Installing CA certificate: {certificate_file}")
            dst_path = "/usr/local/share/ca-certificates/" + certificate_file.name
            self._push_file(source=str(certificate_file), destination=dst_path)

        if certificate_files:
            self._run(["update-ca-certificates"])

    def _check_environment_needs_cleaning(self) -> bool:
        info = self._load_info()
        provider_base = info.get("base")
        built_by = info.get("created-by-snapcraft-version")
        build_base = self.project._get_build_base()

        if provider_base is None or built_by is None:
            self.echoer.warning(
                "Build environment is in unknown state, cleaning first."
            )
            return True
        elif build_base != provider_base:
            self.echoer.warning(
                f"Project base changed from {provider_base!r} to {build_base!r}, cleaning first."
            )
            return True
        elif pkg_resources.parse_version(
            snapcraft._get_version()
        ) < pkg_resources.parse_version(built_by):
            self.echoer.warning(
                f"Build environment was created with newer snapcraft version {built_by!r}, cleaning first."
            )
            return True

        instance_project_dir = info.get("host-project-directory")
        if (
            instance_project_dir is not None
            and self.project._project_dir != instance_project_dir
        ):
            self.echoer.warning(
                f"Build environment project directory changed from {instance_project_dir!r}, cleaning first."
            )
            return True

        return False

    def _ensure_compatible_build_environment(self) -> None:
        """Force clean of build-environment if project is not compatible."""

        if self._check_environment_needs_cleaning():
            self.clean_project()

    def _install_file(self, *, path: str, content: str, permissions: str) -> None:
        basename = os.path.basename(path)

        # Push to a location that can be written to by all backends
        # with unique files depending on path.
        # The path should be a valid path for the target.
        remote_file = "/var/tmp/{}".format(base64.b64encode(path.encode()).decode())

        # Windows cannot open the same file twice, so write to a temporary file that
        # would later be deleted manually.
        with tempfile.NamedTemporaryFile(delete=False, suffix=basename) as temp_file:
            temp_file.write(content.encode())
            temp_file.flush()
            temp_file_path = temp_file.name

        try:
            self._push_file(source=temp_file.name, destination=remote_file)
            self._run(["mv", remote_file, path])

            # This chown is not necessarily needed. but does keep things
            # consistent.
            self._run(["chown", "root:root", path])
            self._run(["chmod", permissions, path])
        finally:
            os.unlink(temp_file_path)

    def _get_code_name_from_build_base(self):
        build_base = self.project._get_build_base()

        return {
            "core": "xenial",
            "core16": "xenial",
            "core18": "bionic",
            "core20": "focal",
        }[build_base]

    def _get_primary_mirror(self) -> str:
        primary_mirror = os.getenv("SNAPCRAFT_BUILD_ENVIRONMENT_PRIMARY_MIRROR", None)

        if primary_mirror is None:
            if platform.machine() in ["AMD64", "i686", "x86_64"]:
                primary_mirror = "http://archive.ubuntu.com/ubuntu"
            else:
                primary_mirror = "http://ports.ubuntu.com/ubuntu-ports"

        return primary_mirror

    def _get_security_mirror(self) -> str:
        security_mirror = os.getenv("SNAPCRAFT_BUILD_ENVIRONMENT_PRIMARY_MIRROR", None)

        if security_mirror is None:
            if platform.machine() in ["AMD64", "i686", "x86_64"]:
                security_mirror = "http://security.ubuntu.com/ubuntu"
            else:
                security_mirror = "http://ports.ubuntu.com/ubuntu-ports"

        return security_mirror

    def _setup_environment(self) -> None:
        self._install_file(
            path="/root/.bashrc",
            content=dedent(
                """\
                        #!/bin/bash
                        export PS1="\\h \\$(/bin/_snapcraft_prompt)# "
                        """
            ),
            permissions="0600",
        )

        self._install_file(
            path="/bin/_snapcraft_prompt",
            content=dedent(
                """\
                        #!/bin/bash
                        if [[ "$PWD" =~ ^$HOME.* ]]; then
                            path="${PWD/#$HOME/\\ ..}"
                            if [[ "$path" == " .." ]]; then
                                ps1=""
                            else
                                ps1="$path"
                            fi
                        else
                            ps1="$PWD"
                        fi
                        echo -n $ps1
                        """
            ),
            permissions="0755",
        )

        self._install_file(path="/etc/apt/sources.list", content="", permissions="0644")

        self._install_file(
            path="/etc/apt/sources.list.d/default.sources",
            content=dedent(
                """\
                    Types: deb
                    URIs: {primary_mirror}
                    Suites: {release} {release}-updates
                    Components: main multiverse restricted universe
                    """.format(
                    primary_mirror=self._get_primary_mirror(),
                    release=self._get_code_name_from_build_base(),
                )
            ),
            permissions="0644",
        )

        self._install_file(
            path="/etc/apt/sources.list.d/default-security.sources",
            content=dedent(
                """\
                        Types: deb
                        URIs: {security_mirror}
                        Suites: {release}-security
                        Components: main multiverse restricted universe
                        """.format(
                    security_mirror=self._get_security_mirror(),
                    release=self._get_code_name_from_build_base(),
                )
            ),
            permissions="0644",
        )

        self._install_file(
            path="/etc/apt/apt.conf.d/00-snapcraft",
            content=dedent(
                """\
                    Apt::Install-Recommends "false";
                    """
            ),
            permissions="0644",
        )

    def _setup_snapcraft(self) -> None:
        self._save_info(
            data={
                "base": self.project._get_build_base(),
                "created-by-snapcraft-version": snapcraft._get_version(),
                "host-project-directory": self.project._project_dir,
            }
        )

        registry_filepath = os.path.join(
            self.provider_project_dir, "snap-registry.yaml"
        )

        # We do not want to inject from the host if not running from the snap
        # or if the provider cannot handle snap mounts.
        # This latter problem should go away when API for retrieving snaps
        # through snapd is generally available.
        if self._get_is_snap_injection_capable():
            inject_from_host = common.is_snap()
        else:
            inject_from_host = False

        snap_injector = SnapInjector(
            registry_filepath=registry_filepath,
            runner=self._run,
            file_pusher=self._push_file,
            inject_from_host=inject_from_host,
        )

        # Note that snap injection order is important.
        # If the build base is core, we do not need to inject snapd.
        # Check for None as this build can be driven from a non snappy enabled
        # system, so we may find ourselves in a situation where the base is not
        # set like on OSX or Windows.
        build_base = self.project._get_build_base()
        if build_base is not None and build_base != "core":
            snap_injector.add(snap_name="snapd")

        # Prevent injecting core18 twice.
        if build_base is not None and build_base != "core18":
            snap_injector.add(snap_name=build_base)

        # Inject snapcraft
        snap_injector.add(snap_name="core18")
        snap_injector.add(snap_name="snapcraft")

        snap_injector.apply()

    def _setup_snapd_proxy(self) -> None:
        """Configure snapd proxy settings from http(s)_proxy."""
        http_proxy = self.build_provider_flags.get("http_proxy")
        if http_proxy:
            self._run(["snap", "set", "system", f"proxy.http={http_proxy}"])
        else:
            self._run(["snap", "unset", "system", "proxy.http"])

        https_proxy = self.build_provider_flags.get("https_proxy")
        if https_proxy:
            self._run(["snap", "set", "system", f"proxy.https={https_proxy}"])
        else:
            self._run(["snap", "unset", "system", "proxy.https"])

    def _get_env_command(self) -> Sequence[str]:
        """Get command sequence for `env` with configured flags."""

        env_list = ["env"]

        # Tell Snapcraft it can take ownership of the host.
        env_list.append("SNAPCRAFT_BUILD_ENVIRONMENT=managed-host")

        # Set the HOME directory.
        env_list.append(f"HOME={self._get_home_directory().as_posix()}")

        # Configure SNAPCRAFT_HAS_TTY.
        has_tty = str(sys.stdout.isatty())
        env_list.append(f"SNAPCRAFT_HAS_TTY={has_tty}")

        # Pass through configurable environment variables.
        for key, value in self.build_provider_flags.items():
            # Ignore keys that aren't relevant to managed instances.
            if key in ["SNAPCRAFT_BIND_SSH", "SNAPCRAFT_ADD_CA_CERTIFICATES"]:
                continue

            if not value:
                continue

            # Ensure item is treated as string and append it.
            value = str(value)
            env_list.append(f"{key}={value}")

        return env_list

    def _get_home_directory(self) -> pathlib.Path:
        """Get user's home directory path."""
        return pathlib.Path("/root")

    def _load_info(self) -> Dict[str, Any]:
        filepath = os.path.join(self.provider_project_dir, "project-info.yaml")
        if not os.path.exists(filepath):
            return dict()

        with open(filepath) as info_file:
            return yaml_utils.load(info_file)

    def _log_run(self, command: Sequence[str]) -> None:
        cmd_string = " ".join([shlex.quote(c) for c in command])
        logger.debug(f"Running: {cmd_string}")

    def _save_info(self, *, data: Dict[str, Any]) -> None:
        filepath = os.path.join(self.provider_project_dir, "project-info.yaml")

        dirpath = os.path.dirname(filepath)
        if dirpath:
            os.makedirs(dirpath, exist_ok=True)

        with open(filepath, "w") as info_file:
            yaml_utils.dump(data, stream=info_file)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import pydoc
from unittest import mock

import fixtures
from testtools.matchers import Contains, Equals, StartsWith

from snapcraft.cli._runner import run
from snapcraft.cli.help import _TOPICS
from tests import fixture_setup

from . import CommandBaseTestCase


class HelpCommandBaseTestCase(CommandBaseTestCase):
    def setUp(self):
        super().setUp()
        # pydoc pager guess can fail, for tests we want a plain pager
        # anyway
        p = mock.patch("pydoc.pager", new=pydoc.plainpager)
        p.start()
        self.addCleanup(p.stop)


class HelpCommandTestCase(HelpCommandBaseTestCase):
    def test_topic_and_plugin_not_found_exits_with_tip(self):
        fake_logger = fixtures.FakeLogger(level=logging.ERROR)
        self.useFixture(fake_logger)

        result = self.run_command(["help", "does-not-exist"])

        self.assertThat(result.exit_code, Equals(1))
        self.assertThat(
            result.output, Contains("There is no help topic, plugin or command")
        )

    def test_topic_and_plugin_adds_ellipsis_for_long_arg(self):
        fake_logger = fixtures.FakeLogger(level=logging.ERROR)
        self.useFixture(fake_logger)

        result = self.run_command(["help", "1234567890123"])

        self.assertThat(result.exit_code, Equals(1))
        self.assertThat(result.output, Contains("1234567890..."))

    def test_print_module_help_for_valid_plugin_default_base(self):
        result = self.run_command(["help", "nil"])

        expected = "Displaying help for the 'nil' plugin for 'core20'."
        output = result.output[: len(expected)]
        self.assertThat(
            output,
            Equals(expected),
            "The help message does not start with {!r} but with "
            "{!r} instead".format(expected, output),
        )

    def test_print_module_help_for_valid_plugin_with_base(self):
        for base in ("core", "core18", "core20"):
            result = self.run_command(["help", "nil", "--base", base])

            expected = f"Displaying help for the 'nil' plugin for {base!r}."
            output = result.output[: len(expected)]
            self.expectThat(
                output,
                Equals(expected),
                "The help message does not start with {!r} but with "
                "{!r} instead".format(expected, output),
            )

    def test_print_module_help_for_valid_plugin_snapcraft_yaml(self):
        self.useFixture(
            fixture_setup.SnapcraftYaml(
                self.path,
                base="core18",
                parts={"part1": {"source": ".", "plugin": "nil"}},
            )
        )
        result = self.run_command(["help", "python", "--base", "core18"])

        expected = (
            "Displaying help for the 'python' plugin for 'core18'.\n\n"
            "The python plugin can be used for"
        )
        output = result.output[: len(expected)]
        self.assertThat(
            output,
            Equals(expected),
            "The help message does not start with {!r} but with "
            "{!r} instead".format(expected, output),
        )

    def test_print_module_named_with_dashes_help_for_valid_plugin(self):
        result = self.run_command(["help", "plainbox-provider", "--base", "core18"])

        expected = "Displaying help for the 'plainbox-provider' plugin for 'core18'."
        self.assertThat(result.output, StartsWith(expected))

    def test_show_module_help_with_devel_for_valid_plugin(self):
        result = self.run_command(["help", "nil", "--devel"])

        expected = "Help on module snapcraft.plugins.v2.nil in snapcraft.plugins"
        output = result.output[: len(expected)]

        self.assertThat(
            output,
            Equals(expected),
            "The help message does not start with {!r} but with "
            "{!r} instead".format(expected, output),
        )

    def test_print_topics(self):
        result = self.run_command(["help", "topics"])

        output = result.output.strip().split("\n")
        for t in _TOPICS:
            self.assertTrue(
                t in output, "Missing topic: {!r} in {!r}".format(t, output)
            )

    def test_print_topic_help_for_valid_topic(self):
        result = self.run_command(["help", "sources"])

        expected = "Common 'source' options."
        output = result.output[: len(expected)]
        self.assertThat(
            output,
            Equals(expected),
            "The help message does not start with {!r} but with "
            "{!r} instead".format(expected, output),
        )

    def test_print_generic_help_by_default(self):
        result = self.run_command(["help"])

        self.assertThat(
            result.output, Contains("Snapcraft is a delightful packaging tool.")
        )
        self.assertThat(result.output, Contains("For more help"))

    def test_no_unicode_in_help_strings(self):
        helps = ["topics"]

        for key in _TOPICS.keys():
            helps.append(str(key))

        # Get a list of plugins
        import os
        from pathlib import Path

        import snapcraft.plugins

        for plugin in Path(snapcraft.plugins.__path__[0]).glob("*.py"):
            if os.path.isfile(str(plugin)) and not os.path.basename(
                str(plugin)
            ).startswith("_"):
                helps.append(os.path.basename(str(plugin)[:-3]))

        for key in helps:
            result = self.run_command(["help", key])
            # An UnicodeEncodeError will be raised if the help text has
            # non-ASCII characters.
            result.output.encode("ascii")


class TopicWithDevelTestCase(HelpCommandBaseTestCase):
    def test_print_topic_help_with_devel_for_valid_topic(self):
        expected = {
            "sources": "Help on package snapcraft",
            "plugins": "Help on package snapcraft",
        }

        for topic in _TOPICS:
            result = self.run_command(["help", topic, "--devel"])
            output = result.output[: len(expected[topic])]
            self.assertThat(
                output,
                Equals(expected[topic]),
                "The help message does not start with {!r} but with "
                "{!r} instead".format(expected[topic], output),
            )


class TestHelpForCommand(HelpCommandBaseTestCase):
    def test_help_for_command(self):
        for command in run.commands:
            result = self.run_command(["help", command])
            self.assertThat(result.exit_code, Equals(0))
            # Verify that the first line of help text is correct
            # to ensure no name squatting takes place.
            self.assertThat(
                result.output, Contains(run.commands[command].help.split("\n")[0])
            )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import TYPE_CHECKING, List

from . import typing
from ._statement import Statement

# Don't use circular imports unless type checking
if TYPE_CHECKING:
    from ._processor import GrammarProcessor  # noqa: F401


class CompoundStatement(Statement):
    """Multiple statements that need to be treated as a group."""

    def __init__(
        self,
        *,
        statements: List[Statement],
        body: typing.Grammar,
        processor: "GrammarProcessor",
        call_stack: typing.CallStack = None
    ) -> None:
        """Create an CompoundStatement instance.

        :param list statements: List of compound statements
        :param list body: The body of the clause.
        :param GrammarProcessor process: GrammarProcessor to use for processing
                                         this statement.
        :param list call_stack: Call stack leading to this statement.
        """
        super().__init__(body=body, processor=processor, call_stack=call_stack)

        self.statements = statements

    def _check(self) -> bool:
        """Check if each statement checks True, in order

        :return: True if each statement agrees that they should be processed,
                 False if elses should be processed.
        :rtype: bool
        """
        for statement in self.statements:
            if not statement._check():
                return False

        return True

    def __eq__(self, other) -> bool:
        if type(other) is type(self):
            return self.statements == other.statements

        return False

    def __str__(self) -> str:
        representation = ""
        for statement in self.statements:
            representation += "{!s} ".format(statement)

        return representation.strip()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import subprocess

from testtools.matchers import Equals, FileContains

from tests import integration


class EnvironmentTestCase(integration.TestCase):
    def test_environment(self):
        self.run_snapcraft("stage", "snapcraft-environment")

        part_install_dir = os.path.join(self.path, self.parts_dir, "env", "install")

        test_name = os.path.join(self.stage_dir, "test_name")
        self.assertThat(test_name, FileContains("test-environment"))

        test_version = os.path.join(self.stage_dir, "test_version")
        self.assertThat(test_version, FileContains("0.1"))

        test_stage = os.path.join(self.stage_dir, "test_stage")
        self.assertThat(
            test_stage, FileContains(os.path.join(self.path, self.stage_dir))
        )

        test_part_install = os.path.join(self.path, self.stage_dir, "test_part_install")
        self.assertThat(test_part_install, FileContains(part_install_dir))

    def test_project_environment_within_snapcraft(self):
        """Replace the SNAPCRAFT_PROJECT_.* occurrences in snapcraft.yaml

        If the correct environment SNAPCRAFT_PROJECT values aren't replaced
        this test will fail its pull step."""
        self.run_snapcraft("pull", "snapcraft-key-values")

    def test_snapcraft_stage_env_replacement(self):
        self.run_snapcraft("stage", "stage_env")

    def test_stage_cmake_plugin_with_replace(self):
        """Replace SNAPCRAFT_PART_* in the part's attributes"""
        self.run_snapcraft("stage", "cmake-with-env-var")

        binary_output = subprocess.check_output(
            [os.path.join(self.stage_dir, "bin", "cmake-with-env-var")]
        )
        sourcedir = os.path.join(self.path, self.parts_dir, "cmake-project", "src")
        builddir = os.path.join(self.path, self.parts_dir, "cmake-project", "build")
        installdir = os.path.join(self.path, self.parts_dir, "cmake-project", "install")
        self.assertThat(
            binary_output.decode("utf-8"),
            Equals(
                "I was built with:\n"
                "PART_SRC: {}\n"
                "PART_BUILD: {}\n"
                "PART_INSTALL: {}\n".format(sourcedir, builddir, installdir)
            ),
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import subprocess

from testtools.matchers import Equals

from tests import integration


class BzrSourceTestCase(integration.BzrSourceBaseTestCase):
    def test_pull_bzr_head(self):
        self.copy_project_to_cwd("bzr-head")

        self.init_source_control()
        self.commit('"1"', unchanged=True)
        self.commit('"2"', unchanged=True)
        # test initial branch
        self.run_snapcraft("pull")
        revno = subprocess.check_output(
            ["bzr", "revno", "-r", "-1", "parts/bzr/src"], universal_newlines=True
        ).strip()
        self.assertThat(revno, Equals("2"))
        # test pull doesn't fail
        self.run_snapcraft("pull")
        revno = subprocess.check_output(
            ["bzr", "revno", "-r", "-1", "parts/bzr/src"], universal_newlines=True
        ).strip()
        self.assertThat(revno, Equals("2"))

    def test_pull_bzr_tag(self):
        self.copy_project_to_cwd("bzr-tag")

        self.init_source_control()
        self.commit('"1"', unchanged=True)
        self.commit('"2"', unchanged=True)
        subprocess.check_call(
            ["bzr", "tag", "-r", "1", "initial"], stderr=subprocess.DEVNULL
        )
        # test initial branch
        self.run_snapcraft("pull")
        revno = self.get_revno("parts/bzr/src")
        self.assertThat(revno, Equals("1"))
        # test pull doesn't fail
        self.run_snapcraft("pull")
        revno = self.get_revno("parts/bzr/src")
        self.assertThat(revno, Equals("1"))

    def test_pull_bzr_commit(self):
        self.copy_project_to_cwd("bzr-commit")

        self.init_source_control()
        self.commit('"1"', unchanged=True)
        self.commit('"2"', unchanged=True)
        # test initial branch
        self.run_snapcraft("pull")
        revno = self.get_revno("parts/bzr/src")
        self.assertThat(revno, Equals("1"))
        # test pull doesn't fail
        self.run_snapcraft("pull")
        revno = self.get_revno("parts/bzr/src")
        self.assertThat(revno, Equals("1"))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2016-2021 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import functools
import operator
import os
import stat
import sys
from textwrap import dedent
from typing import Dict, List, Optional, Set, Union

import click
from tabulate import tabulate

import snapcraft
from snapcraft import formatting_utils, storeapi
from snapcraft._store import StoreClientCLI
from snapcraft.storeapi.constants import DEFAULT_SERIES

from . import echo
from ._channel_map import get_tabulated_channel_map
from ._review import review_snap

_MESSAGE_REGISTER_PRIVATE = dedent(
    """\
    Even though this is private snap, you should think carefully about
    the choice of name and make sure you are confident nobody else will
    have a stronger claim to that particular name. If you are unsure
    then we suggest you prefix the name with your developer identity,
    As '$username-yoyodyne-www-site-content'."""
)
_MESSAGE_REGISTER_CONFIRM = dedent(
    """
    We always want to ensure that users get the software they expect
    for a particular name.

    If needed, we will rename snaps to ensure that a particular name
    reflects the software most widely expected by our community.

    For example, most people would expect 'thunderbird' to be published by
    Mozilla. They would also expect to be able to get other snaps of
    Thunderbird as '$username-thunderbird'.

    Would you say that MOST users will expect {!r} to come from
    you, and be the software you intend to publish there?"""
)
_MESSAGE_REGISTER_SUCCESS = "Congrats! You are now the publisher of {!r}."
_MESSAGE_REGISTER_NO = dedent(
    """
    Thank you! {!r} will remain available.

    In the meantime you can register an alternative name."""
)


@click.group()
def storecli():
    """Store commands"""


def _human_readable_acls(store_client: storeapi.StoreClient) -> str:
    acl = store_client.acl()
    snap_names = []
    snap_ids = acl["snap_ids"]

    if snap_ids is not None:
        try:
            for snap_id in snap_ids:
                snap_names.append(store_client.get_snap_name_for_id(snap_id))
        except TypeError:
            raise RuntimeError(f"invalid snap_ids: {snap_ids!r}")
        acl["snap_names"] = snap_names
    else:
        acl["snap_names"] = None

    human_readable_acl: Dict[str, Union[str, List[str], None]] = {
        "expires": str(acl["expires"])
    }

    for key in ("snap_names", "channels", "permissions"):
        human_readable_acl[key] = acl[key]
        if not acl[key]:
            human_readable_acl[key] = "No restriction"

    return dedent(
        """\
        snaps:       {snap_names}
        channels:    {channels}
        permissions: {permissions}
        expires:     {expires}
    """.format(
            **human_readable_acl
        )
    )


@storecli.command()
@click.argument("snap-name", metavar="<snap-name>")
@click.option("--private", is_flag=True, help="Register the snap as a private one")
@click.option("--store", metavar="<store>", help="Store to register with")
@click.option("--yes", is_flag=True)
def register(snap_name, private, store, yes):
    """Register <snap-name> with the store.

    You can use this command to register an available <snap-name> and become
    the publisher for this snap.

    \b
    Examples:
        snapcraft register thunderbird
    """
    if private:
        click.echo(_MESSAGE_REGISTER_PRIVATE.format(snap_name))
    if yes or echo.confirm(_MESSAGE_REGISTER_CONFIRM.format(snap_name)):
        snapcraft.register(snap_name, is_private=private, store_id=store)
        click.echo(_MESSAGE_REGISTER_SUCCESS.format(snap_name))
    else:
        click.echo(_MESSAGE_REGISTER_NO.format(snap_name))


@storecli.command()
@click.option(
    "--release",
    metavar="<channels>",
    help="Optional comma separated list of channels to release <snap-file>",
)
@click.argument(
    "snap-file",
    metavar="<snap-file>",
    type=click.Path(exists=True, readable=True, resolve_path=True, dir_okay=False),
)
def upload(snap_file, release):
    """Upload <snap-file> to the store.

    By passing --release with a comma separated list of channels the snap would
    be released to the selected channels if the store review passes for this
    <snap-file>.

    This operation will block until the store finishes processing this
    <snap-file>.

    If --release is used, the channel map will be displayed after the
    operation takes place.

    \b
    Examples:
        snapcraft upload my-snap_0.1_amd64.snap
        snapcraft upload my-snap_0.2_amd64.snap --release edge
        snapcraft upload my-snap_0.3_amd64.snap --release candidate,beta
    """
    click.echo("Preparing to upload {!r}.".format(os.path.basename(snap_file)))
    if release:
        channel_list = release.split(",")
        click.echo(
            "After uploading, the resulting snap revision will be released to "
            "{} when it passes the Snap Store review."
            "".format(formatting_utils.humanize_list(channel_list, "and"))
        )
    else:
        channel_list = None

    review_snap(snap_file=snap_file)
    snap_name, snap_revision = snapcraft.upload(snap_file, channel_list)

    echo.info("Revision {!r} of {!r} created.".format(snap_revision, snap_name))
    if channel_list:
        store_client_cli = StoreClientCLI()
        snap_channel_map = store_client_cli.get_snap_channel_map(snap_name=snap_name)

        click.echo(
            get_tabulated_channel_map(
                snap_channel_map,
                architectures=snap_channel_map.get_revision(
                    snap_revision
                ).architectures,
            )
        )


@storecli.command("upload-metadata")
@click.option(
    "--force",
    is_flag=True,
    help="Force metadata update to override any possible conflict",
)
@click.argument(
    "snap-file",
    metavar="<snap-file>",
    type=click.Path(exists=True, readable=True, resolve_path=True, dir_okay=False),
)
def upload_metadata(snap_file, force):
    """Upload metadata from <snap-file> to the store.

    The following information will be retrieved from <snap-file> and used
    to update the store:

    \b
    - summary
    - description
    - icon

    If --force is given, it will force the local metadata into the Store,
    ignoring any possible conflict.

    \b
    Examples:
        snapcraft upload-metadata my-snap_0.1_amd64.snap
        snapcraft upload-metadata my-snap_0.1_amd64.snap --force
    """
    click.echo("Uploading metadata from {!r}".format(os.path.basename(snap_file)))
    snapcraft.upload_metadata(snap_file, force)


@storecli.command()
@click.argument("snap-name", metavar="<snap-name>")
@click.argument("revision", metavar="<revision>")
@click.argument("channels", metavar="<channels>")
@click.option(
    "--progressive",
    type=click.IntRange(0, 100),
    default=100,
    metavar="<percentage>",
    help="set a release progression to a certain percentage.",
)
@click.option(
    "--experimental-progressive-releases",
    is_flag=True,
    help="*EXPERIMENTAL* Enables 'progressive releases'.",
    envvar="SNAPCRAFT_EXPERIMENTAL_PROGRESSIVE_RELEASES",
)
def release(
    snap_name,
    revision,
    channels,
    progressive: Optional[int],
    experimental_progressive_releases: bool,
) -> None:
    """Release <snap-name> on <revision> to the selected store <channels>.
    <channels> is a comma separated list of valid channels on the
    store.

    The <revision> must exist on the store, to see available revisions
    run `snapcraft list-revisions <snap_name>`.

    The channel map will be displayed after the operation takes place.
    To see the status map at any other time run `snapcraft status <snap-name>`.

    The format for channels is `[<track>/]<risk>[/<branch>]` where

    \b
        - <track> is used to have long term release channels. It is implicitly
          set to `latest`. If this snap requires one, it can be created by
          request by having a conversation on https://forum.snapcraft.io
          under the *store* category.
        - <risk> is mandatory and can be either `stable`, `candidate`, `beta`
          or `edge`.
        - <branch> is optional and dynamically creates a channel with a
          specific expiration date.

    \b
    Examples:
        snapcraft release my-snap 8 stable
        snapcraft release my-snap 8 stable/my-branch
        snapcraft release my-snap 9 beta,edge
        snapcraft release my-snap 9 lts-channel/stable
        snapcraft release my-snap 9 lts-channel/stable/my-branch
    """
    # If progressive is set to 100, treat it as None.
    if progressive == 100:
        progressive = None

    if progressive is not None and not experimental_progressive_releases:
        raise click.UsageError(
            "--progressive requires --experimental-progressive-releases."
        )
    elif progressive:
        os.environ["SNAPCRAFT_EXPERIMENTAL_PROGRESSIVE_RELEASES"] = "Y"
        echo.warning("*EXPERIMENTAL* progressive releases in use.")

    store_client_cli = StoreClientCLI()
    release_data = store_client_cli.release(
        snap_name=snap_name,
        revision=revision,
        channels=channels.split(","),
        progressive_percentage=progressive,
    )
    snap_channel_map = store_client_cli.get_snap_channel_map(snap_name=snap_name)
    architectures_for_revision = snap_channel_map.get_revision(
        int(revision)
    ).architectures
    tracks = [storeapi.channels.Channel(c).track for c in channels.split(",")]
    click.echo(
        get_tabulated_channel_map(
            snap_channel_map, tracks=tracks, architectures=architectures_for_revision
        )
    )

    opened_channels = release_data.get("opened_channels", [])
    if len(opened_channels) == 1:
        echo.info(f"The {opened_channels[0]!r} channel is now open.")
    elif len(opened_channels) > 1:
        channels = ("{!r}".format(channel) for channel in opened_channels[:-1])
        echo.info(
            "The {} and {!r} channels are now open.".format(
                ", ".join(channels), opened_channels[-1]
            )
        )


@storecli.command()
@click.argument("snap-name", metavar="<snap-name>")
@click.option(
    "--from-channel",
    metavar="<from-channel>",
    required=True,
    help="The channel to promote from.",
)
@click.option(
    "--to-channel",
    metavar="<to-channel>",
    required=True,
    help="The channel to promote to.",
)
@click.option("--yes", is_flag=True, help="Do not prompt for confirmation.")
def promote(snap_name, from_channel, to_channel, yes):
    """Promote a build set from to a channel.

    A build set is a set of commonly tagged revisions, the most simple
    form of a build set is a set of revisions released to a channel.

    Currently, only channels are supported to release from (<from-channel>)

    Prior to releasing, visual confirmation shall be required.

    The format for channels is `[<track>/]<risk>[/<branch>]` where

    \b
        - <track> is used to have long term release channels. It is implicitly
          set to the default.
        - <risk> is mandatory and can be either `stable`, `candidate`, `beta`
          or `edge`.
        - <branch> is optional and dynamically creates a channel with a
          specific expiration date.

    \b
    Examples:
        snapcraft promote my-snap --from-channel candidate --to-channel stable
        snapcraft promote my-snap --from-channel lts/candidate --to-channel lts/stable
        snapcraft promote my-snap --from-channel stable/patch --to-channel stable
        snapcraft promote my-snap --from-channel experimental/stable --to-channel stable
    """
    echo.warning(
        "snapcraft promote does not have a stable CLI interface. Use with caution in scripts."
    )
    parsed_from_channel = storeapi.channels.Channel(from_channel)
    parsed_to_channel = storeapi.channels.Channel(to_channel)

    if parsed_from_channel == parsed_to_channel:
        raise click.BadOptionUsage(
            "--to-channel", "--from-channel and --to-channel cannot be the same."
        )
    elif (
        parsed_from_channel.risk == "edge"
        and parsed_from_channel.branch is None
        and yes
    ):
        raise click.BadOptionUsage(
            "--from-channel",
            "{!r} is not a valid set value for --from-channel when using --yes.".format(
                parsed_from_channel
            ),
        )

    store = storeapi.StoreClient()
    status_payload = store.get_snap_status(snap_name)

    snap_status = storeapi.status.SnapStatus(
        snap_name=snap_name, payload=status_payload
    )
    from_channel_set = snap_status.get_channel_set(parsed_from_channel)
    echo.info("Build set information for {!r}".format(parsed_from_channel))
    click.echo(
        tabulate(
            sorted(from_channel_set, key=operator.attrgetter("arch")),
            headers=["Arch", "Revision", "Version"],
            tablefmt="plain",
        )
    )
    if yes or echo.confirm(
        "Do you want to promote the current set to the {!r} channel?".format(
            parsed_to_channel
        )
    ):
        for c in from_channel_set:
            store.release(
                snap_name=snap_name,
                revision=str(c.revision),
                channels=[str(parsed_to_channel)],
            )
        snap_channel_map = store.get_snap_channel_map(snap_name=snap_name)
        existing_architectures = snap_channel_map.get_existing_architectures()
        click.echo(
            get_tabulated_channel_map(
                snap_channel_map,
                tracks=[parsed_to_channel.track],
                architectures=existing_architectures,
            )
        )
    else:
        echo.wrapped("Channel promotion cancelled")


@storecli.command()
@click.argument("snap-name", metavar="<snap-name>")
@click.argument("channels", metavar="<channel>...", nargs=-1)
def close(snap_name, channels):
    """Close <channel> for <snap-name>.
    Closing a channel allows the <channel> that is closed to track the channel
    that follows it in the channel release chain. As such closing the
    'candidate' channel would make it track the 'stable' channel.

    The channel map will be displayed after the operation takes place.

    \b
    Examples:
        snapcraft close my-snap beta
        snapcraft close my-snap beta edge
    """
    store = storeapi.StoreClient()
    account_info = store.get_account_information()

    try:
        snap_id = account_info["snaps"][DEFAULT_SERIES][snap_name]["snap-id"]
    except KeyError:
        raise storeapi.errors.StoreChannelClosingPermissionError(
            snap_name, DEFAULT_SERIES
        )

    # Returned closed_channels cannot be trusted as it returns risks.
    store.close_channels(snap_id=snap_id, channel_names=channels)
    if len(channels) == 1:
        msg = "The {} channel is now closed.".format(channels[0])
    else:
        msg = "The {} and {} channels are now closed.".format(
            ", ".join(channels[:-1]), channels[-1]
        )

    snap_channel_map = store.get_snap_channel_map(snap_name=snap_name)
    if snap_channel_map.channel_map:
        closed_tracks = {storeapi.channels.Channel(c).track for c in channels}
        existing_architectures = snap_channel_map.get_existing_architectures()

        click.echo(
            get_tabulated_channel_map(
                snap_channel_map,
                architectures=existing_architectures,
                tracks=closed_tracks,
            )
        )
        click.echo()

    echo.info(msg)


@storecli.command()
@click.option(
    "--experimental-progressive-releases",
    is_flag=True,
    help="*EXPERIMENTAL* Enables 'progressive releases'.",
    envvar="SNAPCRAFT_EXPERIMENTAL_PROGRESSIVE_RELEASES",
)
@click.option(
    "architectures",
    "--arch",
    metavar="<arch>",
    multiple=True,
    help="Limit status to these architectures (can specify multiple times)",
)
@click.option(
    "tracks",
    "--track",
    multiple=True,
    metavar="<track>",
    help="Limit status to these tracks (can specify multiple times)",
)
@click.argument("snap-name", metavar="<snap-name>")
def status(snap_name, architectures, tracks, experimental_progressive_releases):
    """Get the status on the store for <snap-name>.

    \b
    Examples:
        snapcraft status my-snap
        snapcraft status --track 20 my-snap
        snapcraft status --arch amd64 my-snap
    """
    if experimental_progressive_releases:
        os.environ["SNAPCRAFT_EXPERIMENTAL_PROGRESSIVE_RELEASES"] = "Y"
        echo.warning("*EXPERIMENTAL* progressive releases in use.")

    snap_channel_map = StoreClientCLI().get_snap_channel_map(snap_name=snap_name)
    existing_architectures = snap_channel_map.get_existing_architectures()

    if not snap_channel_map.channel_map:
        echo.warning("This snap has no released revisions.")
    else:
        if architectures:
            architectures = set(architectures)
            for architecture in architectures.copy():
                if architecture not in existing_architectures:
                    echo.warning(f"No revisions for architecture {architecture!r}.")
                    architectures.remove(architecture)

            # If we have no revisions for any of the architectures requested, there's
            # nothing to do here.
            if not architectures:
                return
        else:
            architectures = existing_architectures

        if tracks:
            tracks = set(tracks)
            existing_tracks = {
                s.track for s in snap_channel_map.snap.channels if s.track in tracks
            }
            for track in tracks - existing_tracks:
                echo.warning(f"No revisions in track {track!r}.")
            tracks = existing_tracks

            # If we have no revisions in any of the tracks requested, there's
            # nothing to do here.
            if not tracks:
                return
        else:
            tracks = None

        click.echo(
            get_tabulated_channel_map(
                snap_channel_map, architectures=architectures, tracks=tracks
            )
        )


@storecli.command("list-revisions")
@click.option(
    "--arch", metavar="<arch>", help="The snap architecture to get the status for"
)
@click.argument("snap-name", metavar="<snap-name>")
def list_revisions(snap_name, arch):
    """Get the history on the store for <snap-name>.

    This command has an alias of `revisions`.

    \b
    Examples:
        snapcraft list-revisions my-snap
        snapcraft list-revisions my-snap --arch armhf
        snapcraft revisions my-snap
    """
    releases = StoreClientCLI().get_snap_releases(snap_name=snap_name)

    def get_channels_for_revision(revision: int) -> List[str]:
        # channels: the set of channels revision was released to, active or not.
        channels: Set[str] = set()
        # seen_channel: applies to channels regardless of revision.
        # The first channel that shows up for each architecture is to
        # be marked as the active channel, all others are historic.
        seen_channel: Dict[str, Set[str]] = dict()

        for release in releases.releases:
            if release.architecture not in seen_channel:
                seen_channel[release.architecture] = set()

            # If the revision is in this release entry and was not seen
            # before it means that this channel is active and needs to
            # be represented with a *.
            if (
                release.revision == revision
                and release.channel not in seen_channel[release.architecture]
            ):
                channels.add(f"{release.channel}*")
            # All other releases found for a revision are inactive.
            elif (
                release.revision == revision
                and release.channel not in channels
                and f"{release.channel}*" not in channels
            ):
                channels.add(release.channel)

            seen_channel[release.architecture].add(release.channel)

        return sorted(list(channels))

    parsed_revisions = list()
    for rev in releases.revisions:
        if arch and arch not in rev.architectures:
            continue
        channels_for_revision = get_channels_for_revision(rev.revision)
        if channels_for_revision:
            channels = ",".join(channels_for_revision)
        else:
            channels = "-"
        parsed_revisions.append(
            (
                rev.revision,
                rev.created_at,
                ",".join(rev.architectures),
                rev.version,
                channels,
            )
        )

    tabulated_revisions = tabulate(
        parsed_revisions,
        numalign="left",
        headers=["Rev.", "Uploaded", "Arches", "Version", "Channels"],
        tablefmt="plain",
    )

    # 23 revisions + header should not need paging.
    if len(parsed_revisions) < 24:
        click.echo(tabulated_revisions)
    else:
        click.echo_via_pager(tabulated_revisions)


@storecli.command("list")
def list_registered():
    """List snap names registered or shared with you.

    \b
    Examples:
        snapcraft list
    """
    snapcraft.list_registered()


@storecli.command("export-login")
@click.argument(
    "login_file", metavar="FILE", type=click.Path(dir_okay=False, writable=True)
)
@click.option(
    "--snaps", metavar="<snaps>", help="Comma-separated list of snaps to limit access"
)
@click.option(
    "--channels",
    metavar="<channels>",
    help="Comma-separated list of channels to limit access",
)
@click.option(
    "--acls", metavar="<acls>", help="Comma-separated list of ACLs to limit access"
)
@click.option(
    "--expires",
    metavar="<expiration date>",
    help="Date/time (in ISO 8601) when this exported login expires",
)
@click.option(
    "--experimental-login",
    is_flag=True,
    help="*EXPERIMENTAL* Enables login through candid.",
    envvar="SNAPCRAFT_EXPERIMENTAL_LOGIN",
)
def export_login(
    login_file: str,
    snaps: str,
    channels: str,
    acls: str,
    expires: str,
    experimental_login: bool,
):
    """Save login configuration for a store account in FILE.

    This file can then be used to log in to the given account with the
    specified permissions. One can also request the login to be exported to
    stdout instead of a file:

        snapcraft export-login -

    For example, to limit access to the edge channel of any snap the account
    can access:

        snapcraft export-login --channels=edge exported

    Or to limit access to only the edge channel of a single snap:

        snapcraft export-login --snaps=my-snap --channels=edge exported

    To limit access to a single snap, but only until 2019:

        snapcraft export-login --expires="2019-01-01T00:00:00" exported
    """

    snap_list = None
    channel_list = None
    acl_list = None

    if snaps:
        snap_list = []
        for package in snaps.split(","):
            snap_list.append({"name": package, "series": DEFAULT_SERIES})

    if channels:
        channel_list = channels.split(",")

    if acls:
        acl_list = acls.split(",")

    store_client = storeapi.StoreClient(use_candid=experimental_login)
    if store_client.use_candid:
        store_client.login(
            packages=snap_list,
            channels=channel_list,
            acls=acl_list,
            expires=expires,
            save=False,
        )
    else:
        snapcraft.login(
            store=store_client,
            packages=snap_list,
            channels=channel_list,
            acls=acl_list,
            expires=expires,
            save=False,
        )

    # Support a login_file of '-', which indicates a desire to print to stdout
    if login_file.strip() == "-":
        echo.info("\nExported login starts on next line:")
        store_client.export_login(config_fd=sys.stdout, encode=True)
        print()

        preamble = "Login successfully exported and printed above"
        login_action = 'echo "<login>" | snapcraft login --with -'
    else:
        # This is sensitive-- it should only be accessible by the owner
        private_open = functools.partial(os.open, mode=0o600)

        # mypy doesn't have the opener arg in its stub. Ignore its warning
        with open(login_file, "w", opener=private_open) as f:  # type: ignore
            store_client.export_login(config_fd=f)

        # Now that the file has been written, we can just make it
        # owner-readable
        os.chmod(login_file, stat.S_IRUSR)

        preamble = "Login successfully exported to {0!r}".format(login_file)
        login_action = "snapcraft login --with {0}".format(login_file)

    print()
    echo.info(
        dedent(
            """\
        {}. This can now be used with

            {}

        """.format(
                preamble, login_action
            )
        )
    )
    try:
        human_acls = _human_readable_acls(store_client)
        echo.info(
            "to log in to this account with no password and have these "
            f"capabilities:\n{human_acls}"
        )
    except NotImplementedError:
        pass

    echo.warning(
        "This exported login is not encrypted. Do not commit it to version control!"
    )


@storecli.command()
@click.option(
    "--with",
    "login_file",
    metavar="<login file>",
    type=click.File("r"),
    help="Path to file created with 'snapcraft export-login'",
)
@click.option(
    "--experimental-login",
    is_flag=True,
    help="*EXPERIMENTAL* Enables login through candid.",
    envvar="SNAPCRAFT_EXPERIMENTAL_LOGIN",
)
def login(login_file, experimental_login: bool):
    """Login with your Ubuntu One e-mail address and password.

    If you do not have an Ubuntu One account, you can create one at
    https://snapcraft.io/account
    """
    store_client = storeapi.StoreClient(use_candid=experimental_login)
    if store_client.use_candid:
        store_client.login(config_fd=login_file, save=True)
    else:
        snapcraft.login(store=store_client, config_fd=login_file)

    print()

    if login_file:
        try:
            human_acls = _human_readable_acls(store_client)
            echo.info("Login successful. You now have these capabilities:\n")
            echo.info(human_acls)
        except NotImplementedError:
            echo.info("Login successful.")
    else:
        echo.info("Login successful.")


@storecli.command()
def logout():
    """Clear session credentials."""
    store = storeapi.StoreClient()
    store.logout()
    echo.info("Credentials cleared.")


@storecli.command()
def whoami():
    """Returns your login information relevant to the store."""
    account = StoreClientCLI().whoami().account

    click.echo(
        dedent(
            f"""\
        email:        {account.email}
        developer-id: {account.account_id}"""
        )
    )


@storecli.command()
@click.argument("snap-name", metavar="<snap-name>")
@click.argument("track_name", metavar="<track>")
def set_default_track(snap_name: str, track_name: str):
    """Set the default track for <snap-name> to <track>.

    The track must be a valid active track for this operation to be successful.
    """
    store_client_cli = StoreClientCLI()

    # Client-side check to verify that the selected track exists.
    snap_channel_map = store_client_cli.get_snap_channel_map(snap_name=snap_name)
    active_tracks = [
        track.name
        for track in snap_channel_map.snap.tracks
        if track.status in ("default", "active")
    ]
    if track_name not in active_tracks:
        echo.exit_error(
            brief=f"The specified track {track_name!r} does not exist for {snap_name!r}.",
            resolution=f"Ensure the {track_name!r} track exists for the {snap_name!r} snap and try again.",
            details="Valid tracks for {!r}: {}.".format(
                snap_name, ", ".join([f"{t!r}" for t in active_tracks])
            ),
        )

    metadata = dict(default_track=track_name)
    store_client_cli.upload_metadata(snap_name=snap_name, metadata=metadata, force=True)

    echo.info(f"Default track for {snap_name!r} set to {track_name!r}.")


@storecli.command()
@click.argument("snap-name", metavar="<snap-name>")
def list_tracks(snap_name: str) -> None:
    """List channel tracks for <snap-name>.

    This command has an alias of `tracks`.

    Track status, creation dates and version patterns are returned alongside
    the track names in a space formatted table.

    Possible Status values are:

    \b
    - active, visible tracks available for installation
    - default, the default track to install from when not explicit
    - hidden, tracks available for installation but unlisted
    - closed, tracks that are no longer available to install from

    A version pattern is a regular expression that restricts a snap revision
    from being released to a track if the version string set does not match.
    """
    store_client_cli = StoreClientCLI()
    snap_channel_map = store_client_cli.get_snap_channel_map(snap_name=snap_name)

    # Iterate over the entries, replace None with - for consistent presentation
    track_table: List[List[str]] = [
        [
            track.name,
            track.status,
            track.creation_date if track.creation_date else "-",
            track.version_pattern if track.version_pattern else "-",
        ]
        for track in snap_channel_map.snap.tracks
    ]

    click.echo(
        tabulate(
            # Sort by "creation-date".
            sorted(track_table, key=operator.itemgetter(2)),
            headers=["Name", "Status", "Creation-Date", "Version-Pattern"],
            tablefmt="plain",
        )
    )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from snapcraft.internal.remote_build import errors


class TestSnapcraftException:

    scenarios = (
        (
            "LaunchpadGitPushError",
            {
                "exception_class": errors.LaunchpadGitPushError,
                "kwargs": dict(command="foo", exit_code=4),
                "expected_brief": "Failed to push sources to Launchpad.",
                "expected_resolution": "Verify connectivity to https://git.launchpad.net and retry build.",
                "expected_details": "Command 'foo' failed with exit code 4.",
                "expected_docs_url": None,
                "expected_reportable": False,
            },
        ),
        (
            "LaunchpadHttpsError",
            {
                "exception_class": errors.LaunchpadHttpsError,
                "kwargs": dict(),
                "expected_brief": "Failed to connect to Launchpad API service.",
                "expected_resolution": "Verify connectivity to https://api.launchpad.net and retry build.",
                "expected_details": None,
                "expected_docs_url": None,
                "expected_reportable": False,
            },
        ),
    )

    def test_snapcraft_exception_handling(
        self,
        exception_class,
        expected_brief,
        expected_details,
        expected_docs_url,
        expected_reportable,
        expected_resolution,
        kwargs,
    ):
        exception = exception_class(**kwargs)

        assert exception.get_brief() == expected_brief
        assert exception.get_resolution() == expected_resolution
        assert exception.get_details() == expected_details
        assert exception.get_docs_url() == expected_docs_url
        assert exception.get_reportable() == expected_reportable



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from unittest import mock

import fixtures
from testtools.matchers import Contains, Equals, FileExists

import snapcraft.storeapi.errors

from . import StoreCommandsBaseTestCase


class ValidateCommandTestCase(StoreCommandsBaseTestCase):
    def setUp(self):
        super().setUp()

        patcher = mock.patch("snapcraft._store.Popen")
        self.popen_mock = patcher.start()
        rv_mock = mock.Mock()
        rv_mock.returncode = 0
        rv_mock.communicate.return_value = [b"foo", b""]
        self.popen_mock.return_value = rv_mock
        self.addCleanup(patcher.stop)

        self.client.login(email="dummy", password="test correct password")

    def test_validate_success(self):
        result = self.run_command(["validate", "core", "core=3", "test-snap=4"])

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            result.output, Contains("Signing validations assertion for core=3")
        )
        self.assertThat(
            result.output, Contains("Signing validations assertion for test-snap=4")
        )

    def test_validate_with_key(self):
        result = self.run_command(
            ["validate", "core", "core=3", "test-snap=4", "--key-name=keyname"]
        )

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            result.output, Contains("Signing validations assertion for core=3")
        )
        self.assertThat(
            result.output, Contains("Signing validations assertion for test-snap=4")
        )
        self.popen_mock.assert_called_with(
            ["snap", "sign", "-k", "keyname"], stderr=-1, stdin=-1, stdout=-1
        )

    def test_validate_from_branded_store(self):
        # Validating snaps from a branded store requires setting
        # `SNAPCRAFT_UBUNTU_STORE` environment variable to the store 'slug'.
        self.useFixture(
            fixtures.EnvironmentVariable("SNAPCRAFT_UBUNTU_STORE", "Test-Branded")
        )

        result = self.run_command(["validate", "core", "test-snap-branded-store=1"])

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            result.output,
            Contains("Signing validations assertion for test-snap-branded-store=1"),
        )

    def test_validate_unknown_snap(self):
        raised = self.assertRaises(
            snapcraft.storeapi.errors.SnapNotFoundError,
            self.run_command,
            ["validate", "notfound", "core=3", "test-snap=4"],
        )

        self.assertThat(str(raised), Equals("Snap 'notfound' was not found."))

    def test_validate_bad_argument(self):
        raised = self.assertRaises(
            snapcraft.storeapi.errors.InvalidValidationRequestsError,
            self.run_command,
            ["validate", "core", "core=foo"],
        )

        self.assertThat(str(raised), Contains("format must be name=revision"))

    def test_validate_with_snap_name(self):
        self.fake_sign = fixtures.MockPatch(
            "snapcraft._store._sign_assertion", return_value=b""
        )
        self.useFixture(self.fake_sign)

        self.run_command(["validate", "core", "test-snap=3"])

        self.assertThat("core-test-snap-r3.assertion", FileExists())
        self.fake_sign.mock.assert_called_once_with(
            "test-snap=3",
            {
                "type": "validation",
                "authority-id": "abcd",
                "series": "16",
                "snap-id": "good",
                "approved-snap-id": "test-snap-snap-id",
                "approved-snap-revision": "3",
                "timestamp": mock.ANY,
                "revoked": "false",
            },
            None,
            "validations",
        )

    def test_revoke(self):
        self.fake_sign = fixtures.MockPatch(
            "snapcraft._store._sign_assertion", return_value=b""
        )
        self.useFixture(self.fake_sign)

        self.run_command(["validate", "core", "snap-1=3", "--revoke"])

        self.fake_sign.mock.assert_called_once_with(
            "snap-1=3",
            {
                "type": "validation",
                "authority-id": "abcd",
                "series": "16",
                "snap-id": "good",
                "approved-snap-id": "snap-id-1",
                "approved-snap-revision": "3",
                "timestamp": mock.ANY,
                "revoked": "true",
                "revision": "1",
            },
            None,
            "validations",
        )

    def test_no_revoke(self):
        self.fake_sign = fixtures.MockPatch(
            "snapcraft._store._sign_assertion", return_value=b""
        )
        self.useFixture(self.fake_sign)

        self.run_command(["validate", "core", "snap-1=3", "--no-revoke"])

        self.fake_sign.mock.assert_called_once_with(
            "snap-1=3",
            {
                "type": "validation",
                "authority-id": "abcd",
                "series": "16",
                "snap-id": "good",
                "approved-snap-id": "snap-id-1",
                "approved-snap-revision": "3",
                "timestamp": mock.ANY,
                "revoked": "false",
                "revision": "1",
            },
            None,
            "validations",
        )

    def test_validate_fallback_to_snap_id(self):
        self.fake_sign = fixtures.MockPatch(
            "snapcraft._store._sign_assertion", return_value=b""
        )
        self.useFixture(self.fake_sign)

        # XXXYYY is an imaginary snap-id.
        self.run_command(["validate", "core", "XXXYYY=3"])

        self.assertThat("core-XXXYYY-r3.assertion", FileExists())
        self.fake_sign.mock.assert_called_once_with(
            "XXXYYY=3",
            {
                "type": "validation",
                "authority-id": "abcd",
                "series": "16",
                "snap-id": "good",
                "approved-snap-id": "XXXYYY",
                "approved-snap-revision": "3",
                "timestamp": mock.ANY,
                "revoked": "false",
            },
            None,
            "validations",
        )

    def test_validate_with_revoke(self):
        self.fake_sign = fixtures.MockPatch(
            "snapcraft._store._sign_assertion", return_value=b""
        )
        self.useFixture(self.fake_sign)

        self.run_command(["validate", "core", "test-snap=3", "--revoke"])

        self.assertThat("core-test-snap-r3.assertion", FileExists())
        self.fake_sign.mock.assert_called_once_with(
            "test-snap=3",
            {
                "type": "validation",
                "authority-id": "abcd",
                "series": "16",
                "snap-id": "good",
                "approved-snap-id": "test-snap-snap-id",
                "approved-snap-revision": "3",
                "timestamp": mock.ANY,
                "revoked": "true",
            },
            None,
            "validations",
        )

    def test_validate_with_no_revoke(self):
        self.fake_sign = fixtures.MockPatch(
            "snapcraft._store._sign_assertion", return_value=b""
        )
        self.useFixture(self.fake_sign)

        self.run_command(["validate", "core", "test-snap=3", "--no-revoke"])

        self.assertThat("core-test-snap-r3.assertion", FileExists())
        self.fake_sign.mock.assert_called_once_with(
            "test-snap=3",
            {
                "type": "validation",
                "authority-id": "abcd",
                "series": "16",
                "snap-id": "good",
                "approved-snap-id": "test-snap-snap-id",
                "approved-snap-revision": "3",
                "timestamp": mock.ANY,
                "revoked": "false",
            },
            None,
            "validations",
        )



import abc

import requests


class Requests(abc.ABC):
    @abc.abstractmethod
    def _request(self, method: str, urlpath: str, **kwargs) -> requests.Response:
        """Request using method against url."""

    def get(self, urlpath: str, **kwargs) -> requests.Response:
        """Perform a POST request with the given arguments.

        The arguments are the same as for the request function,
        namely params and headers.

        :param str url: url to send the request.
        :return Response of the request.
        """
        return self._request("GET", urlpath, **kwargs)

    def post(self, urlpath: str, **kwargs) -> requests.Response:
        """Perform a POST request with the given arguments.

        The arguments are the same as for the request function,
        namely params and headers.

        :param str url: url to send the request.
        :return Response of the request.
        """
        return self._request("POST", urlpath, **kwargs)

    def put(self, urlpath: str, **kwargs) -> requests.Response:
        """Perform a PUT request with the given arguments.

        The arguments are the same as for the request function,
        namely params and headers.

        :param str url: url to send the request.
        :return Response of the request.
        """
        return self._request("PUT", urlpath, **kwargs)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import re
from collections import OrderedDict
from copy import deepcopy
from typing import Any, Dict, List, Optional, Set, Tuple, Type

from snapcraft.internal.meta.errors import SlotValidationError

logger = logging.getLogger(__name__)


class Slot:
    """Representation of a generic snap slot."""

    def __init__(
        self,
        *,
        slot_name: str,
        slot_dict: Optional[Dict[str, Any]] = None,
        use_string_representation: bool = False,
    ) -> None:
        self._slot_name = slot_name

        if slot_dict is None:
            self._slot_dict: Dict[str, Any] = dict()
        else:
            self._slot_dict = slot_dict

        self.use_string_representation = use_string_representation

    @property
    def slot_name(self) -> str:
        """Read-only to ensure consistency with Snap dictionary mappings."""
        return self._slot_name

    def validate(self) -> None:
        """Validate slot, raising exception on error."""

        # Nothing to validate (yet).
        return

    @classmethod
    def from_dict(cls, *, slot_dict: Dict[str, Any], slot_name: str) -> "Slot":
        """Return applicable Slot instance from dictionary properties."""
        interface = slot_dict.get("interface", slot_name)

        # If we explicitly support the type, use it instead.
        slot_class = SLOT_MAPPINGS.get(interface, None)
        if slot_class is not None:
            return slot_class.from_dict(slot_dict=slot_dict, slot_name=slot_name)

        # Handle the general case.
        return Slot(slot_name=slot_name, slot_dict=slot_dict)

    @classmethod
    def from_object(cls, *, slot_object: Any, slot_name: str) -> "Slot":
        if slot_object is None:
            return Slot(slot_name=slot_name)
        elif isinstance(slot_object, str):
            slot = Slot(slot_name=slot_name, use_string_representation=True)
            slot._slot_dict["interface"] = slot_object
            return slot
        elif isinstance(slot_object, dict):
            return Slot.from_dict(slot_dict=slot_object, slot_name=slot_name)

        raise RuntimeError(f"unknown syntax for slot {slot_name!r}: {slot_object!r}")

    def to_yaml_object(self) -> Optional[Dict[str, Any]]:
        # To output string short-form: "slot-name: interface-type"
        if self.use_string_representation:
            return self._slot_dict["interface"]

        # To output shortest-form: "slot-name-is-interface: <empty>"
        if not self._slot_dict:
            return None

        return OrderedDict(deepcopy(self._slot_dict))

    def __repr__(self) -> str:
        return repr(self.__dict__)

    def __str__(self) -> str:
        return str(self.__dict__)


class ContentSlot(Slot):
    """Representation of a snap content slot."""

    def __init__(
        self,
        *,
        slot_name: str,
        content: Optional[str] = None,
        read: List[str] = None,
        write: List[str] = None,
        use_source_key: bool = True,
    ) -> None:
        super().__init__(slot_name=slot_name)

        if read is None:
            self.read: List[str] = list()
        else:
            self.read = read

        if write is None:
            self.write: List[str] = list()
        else:
            self.write = write

        self.use_source_key = use_source_key
        self._content = content

    @property
    def content(self) -> str:
        if self._content:
            return self._content

        # Defaults to slot_name if unspecified.
        return self.slot_name

    @content.setter
    def content(self, content) -> None:
        self._content = content

    @property
    def interface(self) -> str:
        return "content"

    def validate(self) -> None:
        """Validate content slot, raising exception on error."""

        if not self.read and not self.write:
            raise SlotValidationError(
                slot_name=self.slot_name,
                message="`read` or `write` is required for slot",
            )

    def get_content_dirs(self, installed_path: str) -> Set[str]:
        content_dirs: Set[str] = set()

        for path in self.read + self.write:
            # Strip leading "$SNAP" and "/".
            path = re.sub(r"^\$SNAP", "", path)
            path = re.sub(r"^/", "", path)
            path = re.sub(r"^./", "", path)
            content_dirs.add(os.path.join(installed_path, path))

        return content_dirs

    @classmethod
    def from_dict(cls, *, slot_dict: Dict[str, Any], slot_name: str) -> "ContentSlot":
        slot = ContentSlot(slot_name=slot_name, content=slot_dict.get("content"))

        # Content directories may be nested under "source",
        # but they cannot be in both places according to snapd:
        # https://github.com/snapcore/snapd/blob/master/interfaces/builtin/content.go#L81
        if "source" in slot_dict:
            source_data = slot_dict["source"]
            slot.use_source_key = True
        else:
            source_data = slot_dict
            slot.use_source_key = False

        if "read" in source_data:
            slot.read = source_data["read"]

        if "write" in source_data:
            slot.write = source_data["write"]

        return slot

    def to_yaml_object(self) -> Dict[str, Any]:
        props: List[Tuple[str, Any]] = [("interface", self.interface)]

        # Only include content if set explicitly.
        if self._content:
            props.append(("content", self.content))

        if self.use_source_key:
            source = dict()
            if self.read:
                source["read"] = self.read
            if self.write:
                source["write"] = self.write
            props.append(("source", source))
        else:
            if self.read:
                props.append(("read", self.read))
            if self.write:
                props.append(("write", self.write))

        return OrderedDict(props)


class DbusSlot(Slot):
    """Representation of a snap dbus slot."""

    def __init__(self, *, slot_name: str, bus: str, name: str) -> None:
        super().__init__(slot_name=slot_name)

        self.bus = bus
        self.name = name

    @property
    def interface(self) -> str:
        return "dbus"

    @classmethod
    def from_dict(cls, *, slot_dict: Dict[str, Any], slot_name: str) -> "DbusSlot":
        """Instantiate DbusSlot from dict."""

        if slot_dict.get("interface") != "dbus":
            raise SlotValidationError(
                slot_name=slot_name, message="invalid interface for DbusSlot"
            )

        if "bus" not in slot_dict:
            raise SlotValidationError(
                slot_name=slot_name, message="bus required for DbusSlot"
            )

        if "name" not in slot_dict:
            raise SlotValidationError(
                slot_name=slot_name, message="name required for DbusSlot"
            )

        return DbusSlot(
            slot_name=slot_name, bus=slot_dict["bus"], name=slot_dict["name"]
        )

    def validate(self) -> None:
        """Validate dbus slot. Raise exception if invalid."""

        if not self.bus:
            raise SlotValidationError(
                slot_name=self.slot_name,
                message="valid `bus` is required for dbus slot",
            )

        if not self.name:
            raise SlotValidationError(
                slot_name=self.slot_name,
                message="valid `name` is required for dbus slot",
            )

    def to_yaml_object(self) -> Dict[str, Any]:
        """Return dict of dbus slot."""

        return OrderedDict(
            [("interface", self.interface), ("bus", self.bus), ("name", self.name)]
        )


SLOT_MAPPINGS: Dict[str, Type[Slot]] = dict(content=ContentSlot, dbus=DbusSlot)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018, 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from unittest import mock

from testtools.matchers import Equals, HasLength

from snapcraft.internal import errors
from snapcraft.plugins.v1 import make

from . import PluginsV1BaseTestCase


class MakePluginTest(PluginsV1BaseTestCase):
    def setUp(self):
        super().setUp()

        class Options:
            makefile = None
            make_parameters = []
            make_install_var = "DESTDIR"
            disable_parallel = False
            artifacts = []

        self.options = Options()

    def test_schema(self):
        schema = make.MakePlugin.schema()

        properties = schema["properties"]
        self.assertTrue(
            "makefile" in properties, 'Expected "makefile" to be included in properties'
        )
        self.assertTrue(
            "make-parameters" in properties,
            'Expected "make-parameters" to be included in properties',
        )

        self.assertTrue(
            "make-install-var" in properties,
            'Expected "make-install-var" to be included in properties',
        )

        makefile = properties["makefile"]
        self.assertTrue(
            "type" in makefile, 'Expected "type" to be included in "makefile"'
        )

        makefile_type = makefile["type"]
        self.assertThat(
            makefile_type,
            Equals("string"),
            'Expected "makefile" "type" to be "string", but it '
            'was "{}"'.format(makefile_type),
        )

        make_parameters = properties["make-parameters"]
        self.assertTrue(
            "type" in make_parameters,
            'Expected "type" to be included in "make-parameters"',
        )

        make_parameters_type = make_parameters["type"]
        self.assertThat(
            make_parameters_type,
            Equals("array"),
            'Expected "make-parameters" "type" to be "array", but it '
            'was "{}"'.format(make_parameters_type),
        )

        make_install_var = properties["make-install-var"]
        self.assertTrue(
            "type" in make_install_var,
            'Expected "type" to be included in "make-install-var"',
        )

        make_install_var_type = make_install_var["type"]
        self.assertThat(
            make_install_var_type,
            Equals("string"),
            'Expected "make-install-var" "type" to be "string", but it '
            'was "{}"'.format(make_install_var_type),
        )

        make_install_var_default = make_install_var["default"]
        self.assertThat(
            make_install_var_default,
            Equals("DESTDIR"),
            'Expected "make-install-var" "default" to be "DESTDIR", but it '
            'was "{}"'.format(make_install_var_default),
        )

    def test_get_build_properties(self):
        expected_build_properties = [
            "makefile",
            "make-parameters",
            "make-install-var",
            "artifacts",
        ]
        resulting_build_properties = make.MakePlugin.get_build_properties()

        self.assertThat(
            resulting_build_properties, HasLength(len(expected_build_properties))
        )

        for property in expected_build_properties:
            self.assertIn(property, resulting_build_properties)

    @mock.patch.object(make.MakePlugin, "run")
    def test_build(self, run_mock):
        plugin = make.MakePlugin("test-part", self.options, self.project)
        os.makedirs(plugin.sourcedir)

        plugin.build()

        self.assertThat(run_mock.call_count, Equals(2))
        run_mock.assert_has_calls(
            [
                mock.call(["make", "-j2"], env=None),
                mock.call(
                    ["make", "install", "DESTDIR={}".format(plugin.installdir)],
                    env=None,
                ),
            ]
        )

    @mock.patch.object(make.MakePlugin, "run")
    def test_build_disable_parallel(self, run_mock):
        self.options.disable_parallel = True
        plugin = make.MakePlugin("test-part", self.options, self.project)
        os.makedirs(plugin.sourcedir)

        plugin.build()

        self.assertThat(run_mock.call_count, Equals(2))
        run_mock.assert_has_calls(
            [
                mock.call(["make", "-j1"], env=None),
                mock.call(
                    ["make", "install", "DESTDIR={}".format(plugin.installdir)],
                    env=None,
                ),
            ]
        )

    @mock.patch.object(make.MakePlugin, "run")
    def test_build_makefile(self, run_mock):
        self.options.makefile = "makefile.linux"
        plugin = make.MakePlugin("test-part", self.options, self.project)
        os.makedirs(plugin.sourcedir)

        plugin.build()

        self.assertThat(run_mock.call_count, Equals(2))
        run_mock.assert_has_calls(
            [
                mock.call(["make", "-f", "makefile.linux", "-j2"], env=None),
                mock.call(
                    [
                        "make",
                        "-f",
                        "makefile.linux",
                        "install",
                        "DESTDIR={}".format(plugin.installdir),
                    ],
                    env=None,
                ),
            ]
        )

    @mock.patch.object(make.MakePlugin, "run")
    def test_build_install_var(self, run_mock):
        self.options.make_install_var = "PREFIX"
        plugin = make.MakePlugin("test-part", self.options, self.project)
        os.makedirs(plugin.sourcedir)

        plugin.build()

        self.assertThat(run_mock.call_count, Equals(2))
        run_mock.assert_has_calls(
            [
                mock.call(["make", "-j2"], env=None),
                mock.call(
                    ["make", "install", "PREFIX={}".format(plugin.installdir)], env=None
                ),
            ]
        )

    @mock.patch.object(make.MakePlugin, "run")
    def test_build_empty_install_var(self, run_mock):
        self.options.make_install_var = ""
        plugin = make.MakePlugin("test-part", self.options, self.project)
        os.makedirs(plugin.sourcedir)

        plugin.build()

        self.assertThat(run_mock.call_count, Equals(2))
        run_mock.assert_has_calls(
            [
                mock.call(["make", "-j2"], env=None),
                mock.call(["make", "install"], env=None),
            ]
        )

    @mock.patch.object(make.MakePlugin, "run")
    @mock.patch("snapcraft.file_utils.link_or_copy_tree")
    @mock.patch("snapcraft.file_utils.link_or_copy")
    def test_build_artifacts(self, link_or_copy_mock, link_or_copy_tree_mock, run_mock):
        self.options.artifacts = ["dir_artifact", "file_artifact"]
        plugin = make.MakePlugin("test-part", self.options, self.project)

        os.makedirs(os.path.join(plugin.builddir, "dir_artifact"))

        plugin.build()

        self.assertThat(run_mock.call_count, Equals(1))
        run_mock.assert_has_calls([mock.call(["make", "-j2"], env=None)])
        self.assertThat(link_or_copy_mock.call_count, Equals(1))
        link_or_copy_mock.assert_has_calls(
            [
                mock.call(
                    os.path.join(plugin.builddir, "file_artifact"),
                    os.path.join(plugin.installdir, "file_artifact"),
                )
            ]
        )
        self.assertThat(link_or_copy_tree_mock.call_count, Equals(1))
        link_or_copy_tree_mock.assert_has_calls(
            [
                mock.call(
                    os.path.join(plugin.builddir, "dir_artifact"),
                    os.path.join(plugin.installdir, "dir_artifact"),
                )
            ]
        )

    @mock.patch.object(make.MakePlugin, "run")
    def test_make_with_env(self, run_mock):
        plugin = make.MakePlugin("test-part", self.options, self.project)
        os.makedirs(plugin.sourcedir)

        env = {"foo": "bar"}
        plugin.make(env=env)

        self.assertThat(run_mock.call_count, Equals(2))
        run_mock.assert_has_calls(
            [
                mock.call(["make", "-j2"], env=env),
                mock.call(
                    ["make", "install", "DESTDIR={}".format(plugin.installdir)], env=env
                ),
            ]
        )

    @mock.patch.object(make.MakePlugin, "run")
    def test_make_with_parameters(self, run_mock):
        """Ensure the targets comes before the parameters."""
        self.options.make_parameters = ["FOO=BAR"]
        plugin = make.MakePlugin("test-part", self.options, self.project)
        os.makedirs(plugin.sourcedir)

        plugin.make()

        self.assertThat(run_mock.call_count, Equals(2))
        run_mock.assert_has_calls(
            [
                mock.call(["make", "-j2", "FOO=BAR"], env=None),
                mock.call(
                    ["make", "install", "FOO=BAR", f"DESTDIR={plugin.installdir}"],
                    env=None,
                ),
            ]
        )

    def test_unsupported_base(self):
        self.project._snap_meta.base = "unsupported-base"

        raised = self.assertRaises(
            errors.PluginBaseError,
            make.MakePlugin,
            "test-part",
            self.options,
            self.project,
        )

        self.assertThat(raised.part_name, Equals("test-part"))
        self.assertThat(raised.base, Equals("unsupported-base"))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import subprocess
from collections import namedtuple
from textwrap import dedent

import testscenarios
from testtools.matchers import Contains, FileContains, FileExists

from tests import fixture_setup, integration


class AssetTrackingTestCase(integration.TestCase):
    def test_pull(self):
        self.copy_project_to_cwd("asset-tracking")
        stage_version = self.set_stage_package_version(
            "snapcraft.yaml", part="asset-tracking", package="hello"
        )
        build_version = self.set_build_package_version(
            "snapcraft.yaml", part="asset-tracking", package="hello"
        )

        self.run_snapcraft("pull")

        state_file = os.path.join(self.parts_dir, "asset-tracking", "state", "pull")
        self.assertThat(state_file, FileExists())
        # Verify that the correct version of 'hello' is installed
        self.assertThat(
            state_file,
            FileContains(
                matcher=Contains(
                    dedent(
                        """\
                assets:
                  build-packages:
                  - hello={}
                  build-snaps: []
                  source-details: null
                  stage-packages:
                  - hello={}
                """.format(
                            build_version, stage_version
                        )
                    )
                )
            ),
        )

    def test_pull_global_build_packages_are_excluded(self):
        """
        Ensure global build-packages are not included in each part's
        build-packages data.
        """
        self.copy_project_to_cwd("build-package-global")
        self.set_build_package_version(
            os.path.join("snap", "snapcraft.yaml"), part=None, package="haskell-doc"
        )
        self.run_snapcraft("pull")

        state_file = os.path.join(self.parts_dir, "empty-part", "state", "pull")
        self.assertThat(state_file, FileExists())
        self.assertThat(
            state_file,
            FileContains(
                matcher=Contains(
                    dedent(
                        """\
                assets:
                  build-packages: []
                  build-snaps: []
                  source-details: null
                  stage-packages: []
                """
                    )
                )
            ),
        )

    def test_pull_build_package_with_any_architecture(self):
        self.copy_project_to_cwd("build-package")
        self.set_build_package_architecture(
            os.path.join("snap", "snapcraft.yaml"),
            part="hello",
            package="hello",
            architecture="any",
        )
        self.run_snapcraft("pull")

        state_file = os.path.join(self.parts_dir, "hello", "state", "pull")
        self.assertThat(state_file, FileExists())
        self.assertThat(
            state_file,
            FileContains(
                matcher=Contains(
                    dedent(
                        """\
                assets:
                  build-packages:
                  - hello:any
                  build-snaps: []
                  source-details: null
                  stage-packages: []
                """
                    )
                )
            ),
        )

    def test_pull_with_virtual_build_package(self):
        virtual_package = "fortunes-off"
        self.addCleanup(subprocess.call, ["sudo", "apt-get", "remove", virtual_package])
        self.run_snapcraft("pull", "build-virtual-package")

        state_file = os.path.join("snap", ".snapcraft", "state")
        self.assertThat(state_file, FileExists())
        self.assertThat(
            state_file,
            FileContains(
                matcher=Contains(
                    dedent(
                        """\
                  - {}={}
                """.format(
                            virtual_package,
                            integration.get_package_version(
                                virtual_package, self.distro_series, self.deb_arch
                            ),
                        )
                    )
                )
            ),
        )


TestDetail = namedtuple("TestDetail", ["field", "value"])


class GitAssetTrackingTestCase(testscenarios.WithScenarios, integration.TestCase):

    scenarios = [
        (
            "plain",
            dict(
                part_name="git-part",
                source_branch="''",
                source_commit=True,
                source_tag="''",
            ),
        ),
        (
            "branch",
            dict(
                part_name="git-part-branch",
                source_branch="test-branch",
                source_commit=False,
                source_tag="''",
            ),
        ),
        (
            "tag",
            dict(
                part_name="git-part-tag",
                source_branch="''",
                source_commit=False,
                source_tag="feature-tag",
            ),
        ),
    ]

    def test_pull_git(self):
        repo_fixture = fixture_setup.GitRepo()
        self.useFixture(repo_fixture)
        project_dir = "asset-tracking-git"

        source_commit = repo_fixture.commit if self.source_commit else "''"

        self.run_snapcraft(["pull", self.part_name], project_dir)

        state_file = os.path.join(self.parts_dir, self.part_name, "state", "pull")
        self.assertThat(state_file, FileExists())
        # fall back to the commit if no other source option is provided
        # snapcraft.source.Git doesn't allow both a tag and a commit
        self.assertThat(
            state_file,
            FileContains(
                matcher=Contains(
                    dedent(
                        """\
                assets:
                  build-packages: []
                  build-snaps: []
                  source-details:
                    source: git-source
                    source-branch: {source_branch}
                    source-checksum: ''
                    source-commit: {source_commit}
                    source-tag: {source_tag}
                  stage-packages: []
                """.format(
                            source_branch=self.source_branch,
                            source_commit=source_commit,
                            source_tag=self.source_tag,
                        )
                    )
                )
            ),
        )


class BazaarAssetTrackingTestCase(testscenarios.WithScenarios, integration.TestCase):
    scenarios = [
        ("plain", dict(part_name="bzr-part", source_commit=True, source_tag="''")),
        (
            "tag",
            dict(
                part_name="bzr-part-tag", source_commit=False, source_tag="feature-tag"
            ),
        ),
    ]

    def test_pull_bzr(self):
        repo_fixture = fixture_setup.BzrRepo("bzr-source")
        self.useFixture(repo_fixture)
        project_dir = "asset-tracking-bzr"

        source_commit = repo_fixture.commit if self.source_commit else ""

        self.run_snapcraft(["pull", self.part_name], project_dir)

        state_file = os.path.join(self.parts_dir, self.part_name, "state", "pull")
        self.assertThat(state_file, FileExists())
        self.assertThat(
            state_file,
            FileContains(
                matcher=Contains(
                    dedent(
                        """\
                assets:
                  build-packages: []
                  build-snaps: []
                  source-details:
                    source: bzr-source
                    source-branch: null
                    source-commit: '{source_commit}'
                    source-tag: {source_tag}
                  stage-packages: []
                """.format(
                            source_commit=source_commit, source_tag=self.source_tag
                        )
                    )
                )
            ),
        )


class MercurialAssetTrackingTestCase(testscenarios.WithScenarios, integration.TestCase):
    scenarios = [
        ("plain", dict(part_name="hg-part", source_commit=True, source_tag="''")),
        (
            "tag",
            dict(
                part_name="hg-part-tag", source_commit=False, source_tag="feature-tag"
            ),
        ),
    ]

    def test_pull_hg(self):
        repo_fixture = fixture_setup.HgRepo("hg-source")
        self.useFixture(repo_fixture)
        project_dir = "asset-tracking-hg"

        source_commit = repo_fixture.commit if self.source_commit else "''"

        self.run_snapcraft(["pull", self.part_name], project_dir)

        state_file = os.path.join(self.parts_dir, self.part_name, "state", "pull")
        self.assertThat(state_file, FileExists())
        self.assertThat(
            state_file,
            FileContains(
                matcher=Contains(
                    dedent(
                        """\
                assets:
                  build-packages: []
                  build-snaps: []
                  source-details:
                    source: hg-source
                    source-branch: ''
                    source-commit: {source_commit}
                    source-tag: {source_tag}
                  stage-packages: []
                """.format(
                            source_commit=source_commit, source_tag=self.source_tag
                        )
                    )
                )
            ),
        )


class SubversionAssetTrackingTestCase(integration.TestCase):
    def test_pull_svn(self):
        repo_fixture = fixture_setup.SvnRepo("svn-source")
        self.useFixture(repo_fixture)
        project_dir = "asset-tracking-svn"
        part_name = "svn-part"

        self.run_snapcraft(["pull", part_name], project_dir)

        state_file = os.path.join(self.parts_dir, part_name, "state", "pull")
        self.assertThat(state_file, FileExists())
        self.assertThat(state_file, FileExists())
        self.assertThat(
            state_file,
            FileContains(
                matcher=Contains(
                    dedent(
                        """\
                assets:
                  build-packages: []
                  build-snaps: []
                  source-details:
                    source: svn-source
                    source-branch: null
                    source-commit: '{source_commit}'
                    source-tag: null
                  stage-packages: []
                """.format(
                            source_commit=repo_fixture.commit
                        )
                    )
                )
            ),
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The go plugin can be used for go projects using go.mod.

This plugin uses the common plugin keywords as well as those for "sources".
For more information check the 'plugins' topic for the former and the
'sources' topic for the latter.

Additionally, this plugin uses the following plugin-specific keywords:

    - go-channel
      (string, default: latest/stable)
      The Snap Store channel to install go from.

    - go-buildtags
      (list of strings)
      Tags to use during the go build. Default is not to use any build tags.
"""

from typing import Any, Dict, List, Set

from snapcraft.plugins.v2 import PluginV2


class GoPlugin(PluginV2):
    @classmethod
    def get_schema(cls) -> Dict[str, Any]:
        return {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "type": "object",
            "additionalProperties": False,
            "properties": {
                "go-channel": {"type": "string", "default": "latest/stable"},
                "go-buildtags": {
                    "type": "array",
                    "uniqueItems": True,
                    "items": {"type": "string"},
                    "default": [],
                },
            },
            "required": ["source"],
        }

    def get_build_snaps(self) -> Set[str]:
        return {f"go/{self.options.go_channel}"}

    def get_build_packages(self) -> Set[str]:
        return {"gcc"}

    def get_build_environment(self) -> Dict[str, str]:
        return {
            "SNAPCRAFT_GO_LDFLAGS": "-ldflags -linkmode=external",
            "CGO_ENABLED": "1",
            "GOBIN": "${SNAPCRAFT_PART_INSTALL}/bin",
        }

    def get_build_commands(self) -> List[str]:
        if self.options.go_buildtags:
            tags = "-tags={}".format(",".join(self.options.go_buildtags))
        else:
            tags = ""

        return [
            "go mod download",
            f'go install -p "${{SNAPCRAFT_PARALLEL_BUILD_COUNT}}" {tags} ${{SNAPCRAFT_GO_LDFLAGS}} ./...',
        ]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
import textwrap
from pathlib import Path
from subprocess import CalledProcessError
from unittest import mock
from unittest.mock import call

import fixtures
import pytest
import testtools
from testtools.matchers import Equals

from snapcraft.internal import repo
from snapcraft.internal.repo import errors
from snapcraft.internal.repo.deb_package import DebPackage
from tests import unit


@pytest.fixture(autouse=True)
def mock_env_copy():
    with mock.patch("os.environ.copy", return_value=dict()) as m:
        yield m


class TestPackages(unit.TestCase):
    def setUp(self):
        super().setUp()

        self.fake_apt_cache = self.useFixture(
            fixtures.MockPatch("snapcraft.internal.repo._deb.AptCache")
        ).mock

        self.fake_run = self.useFixture(
            fixtures.MockPatch("subprocess.check_call")
        ).mock

        self.stage_cache_path = Path(self.path, "stage-cache")
        self.debs_path = Path(self.path, "debs")
        self.debs_path.mkdir(parents=True, exist_ok=False)

        repo._deb._DEB_CACHE_DIR = self.debs_path
        repo._deb._STAGE_CACHE_DIR = self.stage_cache_path

        @contextlib.contextmanager
        def fake_tempdir(*, suffix: str, **kwargs):
            temp_dir = Path(self.path, suffix)
            temp_dir.mkdir(exist_ok=True, parents=True)
            yield str(temp_dir)

        self.fake_tmp_mock = self.useFixture(
            fixtures.MockPatch(
                "snapcraft.internal.repo._deb.tempfile.TemporaryDirectory",
                new=fake_tempdir,
            )
        ).mock

        self.stage_packages_path = Path(self.path)

    @mock.patch(
        "snapcraft.internal.repo._deb._DEFAULT_FILTERED_STAGE_PACKAGES",
        {"filtered-pkg-1", "filtered-pkg-2"},
    )
    def test_fetch_stage_packages(self):
        fake_package = self.debs_path / "fake-package_1.0_all.deb"
        fake_package.touch()
        self.fake_apt_cache.return_value.__enter__.return_value.fetch_archives.return_value = [
            ("fake-package", "1.0", fake_package)
        ]

        fetched_packages = repo.Ubuntu.fetch_stage_packages(
            package_names=["fake-package"],
            stage_packages_path=self.stage_packages_path,
            base="core",
            target_arch="amd64",
        )

        self.fake_apt_cache.assert_has_calls(
            [
                call(stage_cache=self.stage_cache_path, stage_cache_arch="amd64"),
                call().__enter__(),
                call().__enter__().update(),
                call().__enter__().mark_packages({"fake-package"}),
                call()
                .__enter__()
                .unmark_packages({"filtered-pkg-1", "filtered-pkg-2"}),
                call().__enter__().fetch_archives(self.debs_path),
            ]
        )

        self.assertThat(fetched_packages, Equals(["fake-package=1.0"]))

    @mock.patch(
        "snapcraft.internal.repo._deb._DEFAULT_FILTERED_STAGE_PACKAGES",
        {"filtered-pkg-1", "filtered-pkg-2", "filtered-pkg-3:amd64", "filtered-pkg-4"},
    )
    def test_fetch_stage_package_filtered_arch_version(self):
        fake_package = self.debs_path / "fake-package_1.0_all.deb"
        fake_package.touch()
        self.fake_apt_cache.return_value.__enter__.return_value.fetch_archives.return_value = [
            ("fake-package", "1.0", fake_package)
        ]

        package_names = [
            "filtered-pkg-1=0.0",
            "filtered-pkg-2:amd64",
            "filtered-pkg-3",
            "non-filtered",
        ]
        fetched_packages = repo.Ubuntu.fetch_stage_packages(
            package_names=package_names,
            stage_packages_path=self.stage_packages_path,
            base="core",
            target_arch="amd64",
        )

        self.fake_apt_cache.assert_has_calls(
            [
                call(stage_cache=self.stage_cache_path, stage_cache_arch="amd64"),
                call().__enter__(),
                call().__enter__().update(),
                call().__enter__().mark_packages(set(package_names)),
                call().__enter__().unmark_packages({"filtered-pkg-4"}),
                call().__enter__().fetch_archives(self.debs_path),
            ]
        )

        self.assertThat(fetched_packages, Equals(["fake-package=1.0"]))

    def test_fetch_virtual_stage_package(self):
        fake_package = self.debs_path / "fake-package_1.0_all.deb"
        fake_package.touch()
        self.fake_apt_cache.return_value.__enter__.return_value.fetch_archives.return_value = [
            ("fake-package", "1.0", fake_package)
        ]

        fetched_packages = repo.Ubuntu.fetch_stage_packages(
            package_names=["virtual-fake-package"],
            stage_packages_path=self.stage_packages_path,
            base="core",
            target_arch="amd64",
        )

        self.assertThat(fetched_packages, Equals(["fake-package=1.0"]))

    def test_fetch_stage_package_with_deps(self):
        fake_package = self.debs_path / "fake-package_1.0_all.deb"
        fake_package.touch()
        fake_package_dep = self.debs_path / "fake-package-dep_1.0_all.deb"
        fake_package_dep.touch()
        self.fake_apt_cache.return_value.__enter__.return_value.fetch_archives.return_value = [
            ("fake-package", "1.0", fake_package),
            ("fake-package-dep", "2.0", fake_package_dep),
        ]

        fetched_packages = repo.Ubuntu.fetch_stage_packages(
            package_names=["fake-package"],
            stage_packages_path=self.stage_packages_path,
            base="core",
            target_arch="amd64",
        )

        self.assertThat(
            fetched_packages,
            Equals(sorted(["fake-package=1.0", "fake-package-dep=2.0"])),
        )

    def test_get_package_fetch_error(self):
        self.fake_apt_cache.return_value.__enter__.return_value.fetch_archives.side_effect = errors.PackageFetchError(
            "foo"
        )

        raised = self.assertRaises(
            errors.PackageFetchError,
            repo.Ubuntu.fetch_stage_packages,
            package_names=["fake-package"],
            stage_packages_path=Path(self.path),
            base="core",
            target_arch="amd64",
        )
        self.assertThat(str(raised), Equals("Package fetch error: foo"))


class BuildPackagesTestCase(unit.TestCase):
    def setUp(self):
        super().setUp()

        self.fake_apt_cache = self.useFixture(
            fixtures.MockPatch("snapcraft.internal.repo._deb.AptCache")
        ).mock

        self.fake_run = self.useFixture(
            fixtures.MockPatch("subprocess.check_call")
        ).mock

        def get_installed_version(package_name, resolve_virtual_packages=False):
            return "1.0" if "installed" in package_name else None

        self.fake_apt_cache.return_value.__enter__.return_value.get_installed_version.side_effect = (
            get_installed_version
        )

        self.fake_is_dumb_terminal = self.useFixture(
            fixtures.MockPatch(
                "snapcraft.repo._deb.is_dumb_terminal", return_value=True
            )
        ).mock

    def test_install_build_package(self):
        self.fake_apt_cache.return_value.__enter__.return_value.get_packages_marked_for_installation.return_value = [
            ("package", "1.0"),
            ("package-installed", "1.0"),
            ("versioned-package", "2.0"),
            ("dependency-package", "1.0"),
        ]

        build_packages = repo.Ubuntu.install_build_packages(
            ["package-installed", "package", "versioned-package=2.0"]
        )

        self.assertThat(
            build_packages,
            Equals(
                [
                    "dependency-package=1.0",
                    "package=1.0",
                    "package-installed=1.0",
                    "versioned-package=2.0",
                ]
            ),
        )
        self.assertThat(
            self.fake_run.mock_calls,
            Equals(
                [
                    call(["sudo", "--preserve-env", "apt-get", "update"]),
                    call(
                        [
                            "sudo",
                            "--preserve-env",
                            "apt-get",
                            "--no-install-recommends",
                            "-y",
                            "--allow-downgrades",
                            "install",
                            "dependency-package=1.0",
                            "package=1.0",
                            "package-installed=1.0",
                            "versioned-package=2.0",
                        ],
                        env={
                            "DEBIAN_FRONTEND": "noninteractive",
                            "DEBCONF_NONINTERACTIVE_SEEN": "true",
                            "DEBIAN_PRIORITY": "critical",
                        },
                    ),
                    call(
                        [
                            "sudo",
                            "apt-mark",
                            "auto",
                            "dependency-package",
                            "package",
                            "package-installed",
                            "versioned-package",
                        ],
                        env={
                            "DEBIAN_FRONTEND": "noninteractive",
                            "DEBCONF_NONINTERACTIVE_SEEN": "true",
                            "DEBIAN_PRIORITY": "critical",
                        },
                    ),
                ]
            ),
        )

    def test_already_installed_no_specified_version(self):
        self.fake_apt_cache.return_value.__enter__.return_value.get_packages_marked_for_installation.return_value = [
            ("package-installed", "1.0")
        ]

        build_packages = repo.Ubuntu.install_build_packages(["package-installed"])

        self.assertThat(build_packages, Equals(["package-installed=1.0"]))
        self.assertThat(self.fake_run.mock_calls, Equals([]))

    def test_already_installed_with_specified_version(self):
        self.fake_apt_cache.return_value.__enter__.return_value.get_packages_marked_for_installation.return_value = [
            ("package-installed", "1.0")
        ]

        build_packages = repo.Ubuntu.install_build_packages(["package-installed=1.0"])

        self.assertThat(build_packages, Equals(["package-installed=1.0"]))
        self.assertThat(self.fake_run.mock_calls, Equals([]))

    def test_already_installed_with_different_version(self):
        self.fake_apt_cache.return_value.__enter__.return_value.get_packages_marked_for_installation.return_value = [
            ("package-installed", "3.0")
        ]

        build_packages = repo.Ubuntu.install_build_packages(["package-installed=3.0"])

        self.assertThat(build_packages, Equals(["package-installed=3.0"]))
        self.assertThat(
            self.fake_run.mock_calls,
            Equals(
                [
                    call(["sudo", "--preserve-env", "apt-get", "update"]),
                    call(
                        [
                            "sudo",
                            "--preserve-env",
                            "apt-get",
                            "--no-install-recommends",
                            "-y",
                            "--allow-downgrades",
                            "install",
                            "package-installed=3.0",
                        ],
                        env={
                            "DEBIAN_FRONTEND": "noninteractive",
                            "DEBCONF_NONINTERACTIVE_SEEN": "true",
                            "DEBIAN_PRIORITY": "critical",
                        },
                    ),
                    call(
                        ["sudo", "apt-mark", "auto", "package-installed"],
                        env={
                            "DEBIAN_FRONTEND": "noninteractive",
                            "DEBCONF_NONINTERACTIVE_SEEN": "true",
                            "DEBIAN_PRIORITY": "critical",
                        },
                    ),
                ]
            ),
        )

    def test_install_virtual_build_package(self):
        self.fake_apt_cache.return_value.__enter__.return_value.get_packages_marked_for_installation.return_value = [
            ("package", "1.0")
        ]

        build_packages = repo.Ubuntu.install_build_packages(["virtual-package"])

        self.assertThat(build_packages, Equals(["package=1.0"]))
        self.assertThat(
            self.fake_run.mock_calls,
            Equals(
                [
                    call(["sudo", "--preserve-env", "apt-get", "update"]),
                    call(
                        [
                            "sudo",
                            "--preserve-env",
                            "apt-get",
                            "--no-install-recommends",
                            "-y",
                            "--allow-downgrades",
                            "install",
                            "package=1.0",
                        ],
                        env={
                            "DEBIAN_FRONTEND": "noninteractive",
                            "DEBCONF_NONINTERACTIVE_SEEN": "true",
                            "DEBIAN_PRIORITY": "critical",
                        },
                    ),
                    call(
                        ["sudo", "apt-mark", "auto", "package"],
                        env={
                            "DEBIAN_FRONTEND": "noninteractive",
                            "DEBCONF_NONINTERACTIVE_SEEN": "true",
                            "DEBIAN_PRIORITY": "critical",
                        },
                    ),
                ]
            ),
        )

    def test_smart_terminal(self):
        self.fake_is_dumb_terminal.return_value = False
        self.fake_apt_cache.return_value.__enter__.return_value.get_packages_marked_for_installation.return_value = [
            ("package", "1.0")
        ]

        repo.Ubuntu.install_build_packages(["package"])

        self.assertThat(
            self.fake_run.mock_calls,
            Equals(
                [
                    call(["sudo", "--preserve-env", "apt-get", "update"]),
                    call(
                        [
                            "sudo",
                            "--preserve-env",
                            "apt-get",
                            "--no-install-recommends",
                            "-y",
                            "--allow-downgrades",
                            "-o",
                            "Dpkg::Progress-Fancy=1",
                            "install",
                            "package=1.0",
                        ],
                        env={
                            "DEBIAN_FRONTEND": "noninteractive",
                            "DEBCONF_NONINTERACTIVE_SEEN": "true",
                            "DEBIAN_PRIORITY": "critical",
                        },
                    ),
                    call(
                        ["sudo", "apt-mark", "auto", "package"],
                        env={
                            "DEBIAN_FRONTEND": "noninteractive",
                            "DEBCONF_NONINTERACTIVE_SEEN": "true",
                            "DEBIAN_PRIORITY": "critical",
                        },
                    ),
                ]
            ),
        )

    def test_invalid_package_requested(self):
        self.fake_apt_cache.return_value.__enter__.return_value.mark_packages.side_effect = errors.PackageNotFoundError(
            "package-invalid"
        )

        self.assertRaises(
            errors.BuildPackageNotFoundError,
            repo.Ubuntu.install_build_packages,
            ["package-invalid"],
        )

    def test_broken_package_apt_install(self):
        self.fake_apt_cache.return_value.__enter__.return_value.get_packages_marked_for_installation.return_value = [
            ("package", "1.0")
        ]
        self.useFixture(
            fixtures.MockPatch(
                "snapcraft.internal.repo._deb.Ubuntu.refresh_build_packages"
            )
        )

        self.fake_run.side_effect = CalledProcessError(100, "apt-get")
        raised = self.assertRaises(
            errors.BuildPackagesNotInstalledError,
            repo.Ubuntu.install_build_packages,
            ["package=1.0"],
        )
        self.assertThat(raised.packages, Equals("package=1.0"))

    def test_refresh_buid_packages(self):
        repo.Ubuntu.refresh_build_packages()

        self.fake_run.assert_called_once_with(
            ["sudo", "--preserve-env", "apt-get", "update"]
        )

    def test_refresh_buid_packages_fails(self):
        self.fake_run.side_effect = CalledProcessError(
            returncode=1, cmd=["sudo", "--preserve-env", "apt-get", "update"]
        )
        self.assertRaises(
            errors.CacheUpdateFailedError, repo.Ubuntu.refresh_build_packages
        )

        self.assertThat(
            self.fake_run.mock_calls,
            Equals([call(["sudo", "--preserve-env", "apt-get", "update"])]),
        )


class PackageForFileTest(unit.TestCase):
    def setUp(self):
        super().setUp()

        def fake_dpkg_query(*args, **kwargs):
            # dpkg-query -S file_path
            if args[0][2] == "/bin/bash":
                return "bash: /bin/bash\n".encode()
            elif args[0][2] == "/bin/sh":
                return (
                    "diversion by dash from: /bin/sh\n"
                    "diversion by dash to: /bin/sh.distrib\n"
                    "dash: /bin/sh\n"
                ).encode()
            else:
                raise CalledProcessError(
                    1,
                    "dpkg-query: no path found matching pattern {}".format(args[0][2]),
                )

        self.useFixture(
            fixtures.MockPatch("subprocess.check_output", side_effect=fake_dpkg_query)
        )

    def test_get_package_for_file(self):
        self.assertThat(repo.Ubuntu.get_package_for_file("/bin/bash"), Equals("bash"))

    def test_get_package_for_file_with_no_leading_slash(self):
        self.assertThat(repo.Ubuntu.get_package_for_file("bin/bash"), Equals("bash"))

    def test_get_package_for_file_with_diversions(self):
        self.assertThat(repo.Ubuntu.get_package_for_file("/bin/sh"), Equals("dash"))

    def test_get_package_for_file_not_found(self):
        self.assertRaises(
            repo.errors.FileProviderNotFound,
            repo.Ubuntu.get_package_for_file,
            "/bin/not-found",
        )


class TestGetPackagesInBase(testtools.TestCase):
    def test_hardcoded_bases(self):
        for base in ("core", "core16", "core18"):
            packages = [
                DebPackage.from_unparsed(p)
                for p in repo._deb._DEFAULT_FILTERED_STAGE_PACKAGES
            ]
            assert repo._deb.get_packages_in_base(base=base) == packages

    @mock.patch.object(repo._deb, "_get_dpkg_list_path")
    def test_package_list_from_dpkg_list(self, mock_dpkg_list_path):
        temp_dir = self.useFixture(fixtures.TempDir())
        dpkg_list_path = Path(f"{temp_dir.path}/dpkg.list")
        mock_dpkg_list_path.return_value = dpkg_list_path
        with dpkg_list_path.open("w") as dpkg_list_file:
            print(
                textwrap.dedent(
                    """\
            Desired=Unknown/Install/Remove/Purge/Hold
            | Status=Not/Inst/Conf-files/Unpacked/halF-conf/Half-inst/trig-aWait/Trig-pend
            |/ Err?=(none)/Reinst-required (Status,Err: uppercase=bad)
            ||/ Name                          Version                    Architecture Description
            +++-=============================-==========================-============-===========
            ii  adduser                       3.118ubuntu1               all          add and rem
            ii  apparmor                      2.13.3-7ubuntu2            amd64        user-space
            ii  apt                           2.0.1                      amd64        commandline
            ii  base-files                    11ubuntu4                  amd64        Debian base
            ii  base-passwd                   3.5.47                     amd64        Debian base
            ii  zlib1g:amd64                  1:1.2.11.dfsg-2ubuntu1     amd64        compression
            """
                ),
                file=dpkg_list_file,
            )

        self.expectThat(
            repo._deb.get_packages_in_base(base="core20"),
            Equals(
                [
                    DebPackage(name="adduser"),
                    DebPackage(name="apparmor"),
                    DebPackage(name="apt"),
                    DebPackage(name="base-files"),
                    DebPackage(name="base-passwd"),
                    DebPackage(name="zlib1g", arch="amd64"),
                ]
            ),
        )

    @mock.patch.object(repo._deb, "_get_dpkg_list_path")
    def test_package_empty_list_from_missing_dpkg_list(self, mock_dpkg_list_path):
        temp_dir = self.useFixture(fixtures.TempDir())
        dpkg_list_path = Path(f"{temp_dir.path}/dpkg.list")
        mock_dpkg_list_path.return_value = dpkg_list_path

        self.expectThat(repo._deb.get_packages_in_base(base="core22"), Equals(list()))


@mock.patch.object(repo._deb, "get_packages_in_base")
def test_get_filtered_stage_package_restricts_core20_ignore_filter(
    mock_get_packages_in_base,
):
    mock_get_packages_in_base.return_value = [
        DebPackage(name="foo"),
        DebPackage(name="foo2"),
        DebPackage(name="python3-attr"),
        DebPackage(name="python3-blinker"),
        DebPackage(name="python3-certifi"),
        DebPackage(name="python3-cffi-backend"),
        DebPackage(name="python3-chardet"),
        DebPackage(name="python3-configobj"),
        DebPackage(name="python3-cryptography"),
        DebPackage(name="python3-idna"),
        DebPackage(name="python3-importlib-metadata"),
        DebPackage(name="python3-jinja2"),
        DebPackage(name="python3-json-pointer"),
        DebPackage(name="python3-jsonpatch"),
        DebPackage(name="python3-jsonschema"),
        DebPackage(name="python3-jwt"),
        DebPackage(name="python3-lib2to3"),
        DebPackage(name="python3-markupsafe"),
        DebPackage(name="python3-more-itertools"),
        DebPackage(name="python3-netifaces"),
        DebPackage(name="python3-oauthlib"),
        DebPackage(name="python3-pyrsistent"),
        DebPackage(name="python3-pyudev"),
        DebPackage(name="python3-requests"),
        DebPackage(name="python3-requests-unixsocket"),
        DebPackage(name="python3-serial"),
        DebPackage(name="python3-six"),
        DebPackage(name="python3-urllib3"),
        DebPackage(name="python3-urwid"),
        DebPackage(name="python3-yaml"),
        DebPackage(name="python3-zipp"),
    ]

    filtered_names = repo._deb._get_filtered_stage_package_names(
        base="core20", package_list=[]
    )

    assert filtered_names == {"foo", "foo2"}


@mock.patch.object(repo._deb, "get_packages_in_base")
def test_get_filtered_stage_package_empty_ignore_filter(mock_get_packages_in_base):
    mock_get_packages_in_base.return_value = [
        DebPackage(name="some-base-pkg"),
        DebPackage(name="some-other-base-pkg"),
    ]

    filtered_names = repo._deb._get_filtered_stage_package_names(
        base="core00", package_list=[]
    )

    assert filtered_names == {"some-base-pkg", "some-other-base-pkg"}



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2021 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
from pathlib import PosixPath
import subprocess
from textwrap import dedent
from unittest import mock

import fixtures
from click.testing import CliRunner

from snapcraft import storeapi
from snapcraft.cli._runner import run
from snapcraft.storeapi.v2.channel_map import ChannelMap
from snapcraft.storeapi.v2.releases import Releases
from tests import fixture_setup, unit

_sample_keys = [
    {
        "name": "default",
        "sha3-384": "vdEeQvRxmZ26npJCFaGnl-VfGz0lU2jZZkWp_s7E-RxVCNtH2_mtjcxq2NkDKkIp",
    },
    {
        "name": "another",
        "sha3-384": "JsfToV5hO2eN9l89pYYCKXUioTERrZIIHUgQQd47jW8YNNBskupiIjWYd3KXLY_D",
    },
]


def get_sample_key(name):
    for key in _sample_keys:
        if key["name"] == name:
            return key
    raise KeyError(name)


original_check_output = subprocess.check_output


def mock_check_output(command, *args, **kwargs):
    if isinstance(command[0], PosixPath):
        command[0] = str(command[0])
    if command[0].endswith("unsquashfs") or command[0].endswith("xdelta3"):
        return original_check_output(command, *args, **kwargs)
    elif command[0].endswith("snap") and command[1:] == ["keys", "--json"]:
        return json.dumps(_sample_keys)
    elif command[0].endswith("snap") and command[1] == "export-key":
        if not command[2].startswith("--account="):
            raise AssertionError("Unhandled command: {}".format(command))
        account_id = command[2][len("--account=") :]
        name = command[3]
        # This isn't a full account-key-request assertion, but it's enough
        # for testing.
        return dedent(
            """\
            type: account-key-request
            account-id: {account_id}
            name: {name}
            public-key-sha3-384: {sha3_384}
            """
        ).format(
            account_id=account_id, name=name, sha3_384=get_sample_key(name)["sha3-384"]
        )
    elif command[0].endswith("snap") and command[1:] == [
        "create-key",
        "new-key",
    ]:
        pass
    else:
        raise AssertionError("Unhandled command: {}".format(command))


class CommandBaseTestCase(unit.TestCase):
    def setUp(self):
        super().setUp()
        self.runner = CliRunner()

    def run_command(self, args, **kwargs):
        # For click testing, runner will overwrite the descriptors for stdio -
        # ensure TTY always appears connected.
        self.useFixture(
            fixtures.MockPatch("snapcraft.cli.echo.is_tty_connected", return_value=True)
        )

        with mock.patch("sys.argv", args):
            return self.runner.invoke(run, args, catch_exceptions=False, **kwargs)


class LifecycleCommandsBaseTestCase(CommandBaseTestCase):
    def setUp(self):
        super().setUp()

        self.useFixture(fixtures.EnvironmentVariable("SNAPCRAFT_BUILD_ENVIRONMENT"))

        self.fake_lifecycle_clean = fixtures.MockPatch(
            "snapcraft.internal.lifecycle.clean"
        )
        self.useFixture(self.fake_lifecycle_clean)

        self.fake_lifecycle_execute = fixtures.MockPatch(
            "snapcraft.internal.lifecycle.execute"
        )
        self.useFixture(self.fake_lifecycle_execute)

        self.fake_pack = fixtures.MockPatch("snapcraft.cli.lifecycle._pack")
        self.useFixture(self.fake_pack)

        self.snapcraft_yaml = fixture_setup.SnapcraftYaml(
            self.path,
            parts={
                "part0": {"plugin": "nil"},
                "part1": {"plugin": "nil"},
                "part2": {"plugin": "nil"},
            },
        )
        self.useFixture(self.snapcraft_yaml)

        self.provider_class_mock = mock.MagicMock()
        self.provider_mock = mock.MagicMock()
        self.provider_class_mock.return_value.__enter__.return_value = (
            self.provider_mock
        )

        self.fake_get_provider_for = fixtures.MockPatch(
            "snapcraft.internal.build_providers.get_provider_for",
            return_value=self.provider_class_mock,
        )
        self.useFixture(self.fake_get_provider_for)

    def assert_clean_not_called(self):
        self.fake_lifecycle_clean.mock.assert_not_called()
        self.provider_mock.clean.assert_not_called()
        self.provider_mock.clean_project.assert_not_called()


class StoreCommandsBaseTestCase(CommandBaseTestCase):
    def setUp(self):
        super().setUp()
        self.fake_store = fixture_setup.FakeStore()
        self.useFixture(self.fake_store)
        self.client = storeapi.StoreClient()


class FakeStoreCommandsBaseTestCase(CommandBaseTestCase):
    def setUp(self):
        super().setUp()

        # Our experimental environment variable is sticky
        self.useFixture(
            fixtures.EnvironmentVariable(
                "SNAPCRAFT_EXPERIMENTAL_PROGRESSIVE_RELEASES", None
            )
        )

        self.fake_store_login = fixtures.MockPatchObject(storeapi.StoreClient, "login")
        self.useFixture(self.fake_store_login)

        self.fake_store_register = fixtures.MockPatchObject(
            storeapi._dashboard_api.DashboardAPI, "register"
        )
        self.useFixture(self.fake_store_register)

        self.fake_store_account_info = fixtures.MockPatchObject(
            storeapi._dashboard_api.DashboardAPI,
            "get_account_information",
            return_value={
                "account_id": "abcd",
                "account_keys": list(),
                "snaps": {
                    "16": {
                        "snap-test": {
                            "snap-id": "snap-test-snap-id",
                            "status": "Approved",
                            "private": False,
                            "since": "2016-12-12T01:01Z",
                            "price": "0",
                        },
                        "basic": {
                            "snap-id": "basic-snap-id",
                            "status": "Approved",
                            "private": False,
                            "since": "2016-12-12T01:01Z",
                            "price": "0",
                        },
                    }
                },
            },
        )
        self.useFixture(self.fake_store_account_info)

        self.fake_store_status = fixtures.MockPatchObject(
            storeapi._dashboard_api.DashboardAPI, "snap_status", return_value=dict()
        )
        self.useFixture(self.fake_store_status)

        self.fake_store_release = fixtures.MockPatchObject(
            storeapi.StoreClient, "release"
        )
        self.useFixture(self.fake_store_release)

        self.fake_store_register_key = fixtures.MockPatchObject(
            storeapi._dashboard_api.DashboardAPI, "register_key"
        )
        self.useFixture(self.fake_store_register_key)

        # channel-map endpoint
        self.channel_map = ChannelMap.unmarshal(
            {
                "channel-map": [
                    {
                        "architecture": "amd64",
                        "channel": "2.1/beta",
                        "expiration-date": None,
                        "revision": 19,
                        "progressive": {
                            "paused": None,
                            "percentage": None,
                            "current-percentage": None,
                        },
                    },
                    {
                        "architecture": "amd64",
                        "channel": "2.0/beta",
                        "expiration-date": None,
                        "revision": 18,
                        "progressive": {
                            "paused": None,
                            "percentage": None,
                            "current-percentage": None,
                        },
                    },
                ],
                "revisions": [
                    {"architectures": ["amd64"], "revision": 19, "version": "10"},
                    {"architectures": ["amd64"], "revision": 18, "version": "10"},
                ],
                "snap": {
                    "name": "snap-test",
                    "channels": [
                        {
                            "branch": None,
                            "fallback": None,
                            "name": "2.1/stable",
                            "risk": "stable",
                            "track": "2.1",
                        },
                        {
                            "branch": None,
                            "fallback": "2.1/stable",
                            "name": "2.1/candidate",
                            "risk": "candidate",
                            "track": "2.1",
                        },
                        {
                            "branch": None,
                            "fallback": "2.1/candidate",
                            "name": "2.1/beta",
                            "risk": "beta",
                            "track": "2.1",
                        },
                        {
                            "branch": None,
                            "fallback": "2.1/beta",
                            "name": "2.1/edge",
                            "risk": "edge",
                            "track": "2.1",
                        },
                        {
                            "branch": None,
                            "fallback": None,
                            "name": "2.0/stable",
                            "risk": "stable",
                            "track": "2.0",
                        },
                        {
                            "branch": None,
                            "fallback": "2.0/stable",
                            "name": "2.0/candidate",
                            "risk": "candidate",
                            "track": "2.0",
                        },
                        {
                            "branch": None,
                            "fallback": "2.0/candidate",
                            "name": "2.0/beta",
                            "risk": "beta",
                            "track": "2.0",
                        },
                        {
                            "branch": None,
                            "fallback": "2.0/beta",
                            "name": "2.0/edge",
                            "risk": "edge",
                            "track": "2.0",
                        },
                    ],
                    "default-track": "2.1",
                    "tracks": [
                        {
                            "name": "2.0",
                            "status": "default",
                            "creation-date": "2019-10-17T14:11:59Z",
                            "version-pattern": "2\\.*",
                        },
                        {
                            "name": "latest",
                            "status": "active",
                            "creation-date": None,
                            "version-pattern": None,
                        },
                    ],
                },
            }
        )
        self.fake_store_get_snap_channel_map = fixtures.MockPatchObject(
            storeapi.StoreClient, "get_snap_channel_map", return_value=self.channel_map
        )
        self.useFixture(self.fake_store_get_snap_channel_map)

        self.releases = Releases.unmarshal(
            {
                "revisions": [
                    {
                        "architectures": ["i386"],
                        "base": "core20",
                        "build_url": None,
                        "confinement": "strict",
                        "created_at": " 2016-09-27T19:23:40Z",
                        "grade": "stable",
                        "revision": 2,
                        "sha3-384": "a9060ef4872ccacbfa440617a76fcd84967896b28d0d1eb7571f00a1098d766e7e93353b084ba6ad841d7b14b95ede48",
                        "size": 20,
                        "status": "Published",
                        "version": "2.0.1",
                    },
                    {
                        "architectures": ["amd64"],
                        "base": "core20",
                        "build_url": None,
                        "confinement": "strict",
                        "created_at": "2016-09-27T18:38:43Z",
                        "grade": "stable",
                        "revision": 1,
                        "sha3-384": "a9060ef4872ccacbfa440617a76fcd84967896b28d0d1eb7571f00a1098d766e7e93353b084ba6ad841d7b14b95ede48",
                        "size": 20,
                        "status": "Published",
                        "version": "2.0.2",
                    },
                ],
                "releases": [
                    {
                        "architecture": "amd64",
                        "branch": None,
                        "channel": "latest/stable",
                        "expiration-date": None,
                        "revision": 1,
                        "risk": "stable",
                        "track": "latest",
                        "when": "2020-02-12T17:51:40.891996Z",
                    },
                    {
                        "architecture": "i386",
                        "branch": None,
                        "channel": "latest/stable",
                        "expiration-date": None,
                        "revision": None,
                        "risk": "stable",
                        "track": "latest",
                        "when": "2020-02-11T17:51:40.891996Z",
                    },
                    {
                        "architecture": "amd64",
                        "branch": None,
                        "channel": "latest/edge",
                        "expiration-date": None,
                        "revision": 1,
                        "risk": "stable",
                        "track": "latest",
                        "when": "2020-01-12T17:51:40.891996Z",
                    },
                ],
            }
        )
        self.fake_store_get_releases = fixtures.MockPatchObject(
            storeapi.StoreClient, "get_snap_releases", return_value=self.releases
        )
        self.useFixture(self.fake_store_get_releases)

        # Uploading
        self.mock_tracker = mock.Mock(storeapi._status_tracker.StatusTracker)
        self.mock_tracker.track.return_value = {
            "code": "ready_to_release",
            "processed": True,
            "can_release": True,
            "url": "/fake/url",
            "revision": 19,
        }
        self.fake_store_upload_precheck = fixtures.MockPatchObject(
            storeapi.StoreClient, "upload_precheck"
        )
        self.useFixture(self.fake_store_upload_precheck)

        self.fake_store_upload = fixtures.MockPatchObject(
            storeapi.StoreClient, "upload", return_value=self.mock_tracker
        )
        self.useFixture(self.fake_store_upload)

        # Mock the snap command, pass through a select few.
        self.fake_check_output = fixtures.MockPatch(
            "subprocess.check_output", side_effect=mock_check_output
        )
        self.useFixture(self.fake_check_output)

        # Pretend that the snap command is available
        self.fake_package_installed = fixtures.MockPatch(
            "snapcraft.internal.repo.Repo.is_package_installed", return_value=True
        )
        self.useFixture(self.fake_package_installed)
